#!/usr/bin/python3
# PYTHON_ARGCOMPLETE_OK

"""
Fetches online ArduPilot parameter documentation (if not cached) locally.

and adds it to the specified file or to all *.param and *.parm files in the specified directory.

1. Checks if a local cache of the apm.pdef.xml file exists in the target directory or on the directory of the target file:
 - If it does, the script loads the file content.
 - If it doesn't, the script sends a GET request to the URL to fetch the XML data for the requested vehicle type.
2. Parses the XML data and creates a dictionary of parameter documentation.
3. DELETES all comments that start at the beginning of a line
4. Adds the parameter documentation to the target file or to all *.param,*.parm files in the target directory.

Supports AP_Periph, AntennaTracker, ArduCopter, ArduPlane, ArduSub, Blimp, Heli, Rover and SITL vehicle types
Supports both Mission Planner and MAVProxy file formats
Supports sorting the parameters

Has unit tests with 88% coverage

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import argparse
import glob
import logging
import re
from os import environ as os_environ
from os import path as os_path
from os import popen as os_popen
from sys import exc_info as sys_exc_info
from sys import exit as sys_exit
from types import TracebackType
from typing import Any, Optional, Union
from xml.etree import ElementTree as ET  # no parsing, just data-structure manipulation

import argcomplete
from argcomplete.completers import FilesCompleter
from defusedxml import ElementTree as DET  # noqa: N814, just parsing, no data-structure manipulation

# URL of the XML file
BASE_URL = "https://autotest.ardupilot.org/Parameters/"

PARAM_DEFINITION_XML_FILE = "apm.pdef.xml"
LUA_PARAM_DEFINITION_XML_FILE = "24_inflight_magnetometer_fit_setup.pdef.xml"

# ArduPilot parameter names start with a capital letter and can have capital letters, numbers and _
PARAM_NAME_REGEX = r"^[A-Z][A-Z_0-9]*"
PARAM_NAME_MAX_LEN = 16
VERSION = "1.0"

# mypy: disable-error-code="unused-ignore"


def create_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Fetches on-line ArduPilot parameter documentation and adds it to the "
        "specified file or to all *.param and *.parm files in the specified directory."
    )
    parser.add_argument(  # type: ignore[attr-defined]
        "target",
        help="The target file or directory.",
    ).completer = FilesCompleter(allowednames=(".param", ".parm"))  # type: ignore[no-untyped-call]
    parser.add_argument(
        "-d",
        "--delete-documentation-annotations",
        action="store_true",
        help="Delete parameter documentation annotations (comments above parameters). Default is %(default)s",
    )
    parser.add_argument(
        "-f",
        "--firmware-version",
        default="latest",
        help="Flight controller firmware version. Default is %(default)s.",
    )
    parser.add_argument(
        "-s",
        "--sort",
        choices=["none", "missionplanner", "mavproxy"],
        default="none",
        help="Sort the parameters in the file. Default is %(default)s.",
    )
    parser.add_argument(
        "-t",
        "--vehicle-type",
        choices=["AP_Periph", "AntennaTracker", "ArduCopter", "ArduPlane", "ArduSub", "Blimp", "Heli", "Rover", "SITL"],
        default="ArduCopter",
        help="The type of the vehicle. Default is %(default)s.",
    )
    parser.add_argument(
        "-m",
        "--max-line-length",
        type=int,
        default=100,
        help="Maximum documentation line length. Default is %(default)s.",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Increase output verbosity, print ReadOnly parameter list. Default is %(default)s.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {VERSION}",
        help="Display version information and exit.",
    )

    argcomplete.autocomplete(parser)
    return parser


def parse_arguments() -> argparse.Namespace:
    parser = create_argument_parser()

    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
    else:
        logging.basicConfig(level=logging.WARNING, format="%(levelname)s: %(message)s")

    # Custom validation for --max-line-length
    def check_max_line_length(value: int) -> int:
        if value < 50 or value > 300:
            logging.critical("--max-line-length must be in the interval 50 .. 300, not %d", value)
            msg = "Correct it and try again"
            raise SystemExit(msg)
        return value

    args.max_line_length = check_max_line_length(args.max_line_length)

    return args


class Par:
    """
    Represents a parameter with a value and an optional comment.

    Attributes:
        value (float): The value of the parameter.
        comment (Optional[str]): An optional comment associated with the parameter.

    """

    def __init__(self, value: float, comment: Optional[str] = None) -> None:
        self.value = value
        self.comment = comment

    def __eq__(self, other: object) -> bool:
        """Equality operation."""
        if isinstance(other, Par):
            return self.value == other.value and self.comment == other.comment
        return False

    @staticmethod
    def load_param_file_into_dict(param_file: str) -> dict[str, "Par"]:
        """
        Loads an ArduPilot parameter file into a dictionary with name, value pairs.

        Args:
            param_file (str): The name of the parameter file to load.

        Returns:
            dict: A dictionary containing the parameters from the file.

        """
        parameter_dict: dict[str, Par] = {}
        try:
            with open(param_file, encoding="utf-8") as f_handle:
                for i, f_line in enumerate(f_handle, start=1):
                    original_line = f_line
                    line = f_line.strip()
                    comment = None
                    if not line:
                        continue  # skip empty lines
                    if line[0] == "#":
                        continue  # skip comments
                    if "#" in line:
                        line, comment = line.split("#", 1)  # strip trailing comments
                        comment = comment.strip()
                    if "," in line:
                        # parse mission planner style parameter files
                        parameter, value = line.split(",", 1)
                    elif " " in line:
                        # parse mavproxy style parameter files
                        parameter, value = line.split(" ", 1)
                    elif "\t" in line:
                        parameter, value = line.split("\t", 1)
                    else:
                        msg = f"Missing parameter-value separator: {line} in {param_file} line {i}"
                        raise SystemExit(msg)
                    parameter = parameter.strip()
                    Par.validate_parameter(param_file, parameter_dict, i, original_line, comment, parameter, value)
        except UnicodeDecodeError as exp:
            msg = f"Fatal error reading {param_file}: {exp}"
            raise SystemExit(msg) from exp
        return parameter_dict

    @staticmethod
    def validate_parameter(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        param_file: str,
        parameter_dict: dict[str, "Par"],
        i: int,
        original_line: str,
        comment: Union[None, str],
        parameter_name: str,
        value: str,
    ) -> None:
        if len(parameter_name) > PARAM_NAME_MAX_LEN:
            msg = f"Too long parameter name: {parameter_name} in {param_file} line {i}"
            raise SystemExit(msg)
        if not re.match(PARAM_NAME_REGEX, parameter_name):
            msg = f"Invalid characters in parameter name {parameter_name} in {param_file} line {i}"
            raise SystemExit(msg)
        if parameter_name in parameter_dict:
            msg = f"Duplicated parameter {parameter_name} in {param_file} line {i}"
            raise SystemExit(msg)
        try:
            fvalue = float(value.strip())
            parameter_dict[parameter_name] = Par(fvalue, comment)
        except ValueError as exc:
            msg = f"Invalid parameter value {value} in {param_file} line {i}"
            raise SystemExit(msg) from exc
        except OSError as exc:
            _exc_type, exc_value, exc_traceback = sys_exc_info()
            if isinstance(exc_traceback, TracebackType):
                fname = os_path.split(exc_traceback.tb_frame.f_code.co_filename)[1]
                logging.critical("in line %s of file %s: %s", exc_traceback.tb_lineno, fname, exc_value)
                msg = f"Caused by line {i} of file {param_file}: {original_line}"
                raise SystemExit(msg) from exc

    @staticmethod
    def missionplanner_sort(item: str) -> tuple[str, ...]:
        """
        Sorts a parameter name according to the rules defined in the Mission Planner software.

        Args:
            item: The parameter name to sort.

        Returns:
            A tuple representing the sorted parameter name.

        """
        parts = item.split("_")  # Split the parameter name by underscore
        # Compare the parts separately
        return tuple(parts)

    @staticmethod
    def format_params(param_dict: dict[str, "Par"], file_format: str = "missionplanner") -> list[str]:  # pylint: disable=too-many-branches
        """
        Formats the parameters in the provided dictionary into a list of strings.

        Each string in the returned list is a formatted representation of a parameter,
        consisting of the parameter's name, its value, and optionally its comment.
        The comment is included if it is present in the parameter's 'Par' object.

        Args:
            param_dict (Dict[str, 'Par']): A dictionary of 'Par' objects.
                                           Each key is a parameter name and each value is a 'Par' object.
                                           Par can be a simple float or a Par object with a comment.
            file_format (str): Can be "missionplanner" or "mavproxy"

        Returns:
            List[str]: A list of strings, each string representing a parameter from the input dictionary
                       in the format "name,value # comment".

        """
        if file_format == "missionplanner":
            param_dict = dict(sorted(param_dict.items(), key=lambda x: Par.missionplanner_sort(x[0])))  # sort alphabetically
        elif file_format == "mavproxy":
            param_dict = dict(sorted(param_dict.items()))  # sort in ASCIIbetical order
        else:
            msg = f"ERROR: Unsupported file format {file_format}"
            raise SystemExit(msg)

        formatted_params = []
        if file_format == "missionplanner":
            for key, parameter in param_dict.items():
                if isinstance(parameter, Par):
                    if parameter.comment:
                        formatted_params.append(
                            f"{key},{format(parameter.value, '.6f').rstrip('0').rstrip('.')}  # {parameter.comment}"
                        )
                    else:
                        formatted_params.append(f"{key},{format(parameter.value, '.6f').rstrip('0').rstrip('.')}")
                else:
                    formatted_params.append(f"{key},{format(parameter, '.6f').rstrip('0').rstrip('.')}")
        elif file_format == "mavproxy":
            for key, parameter in param_dict.items():
                if isinstance(parameter, Par):
                    if parameter.comment:
                        formatted_params.append(f"{key:<16} {parameter.value:<8.6f}  # {parameter.comment}")
                    else:
                        formatted_params.append(f"{key:<16} {parameter.value:<8.6f}")
                else:
                    formatted_params.append(f"{key:<16} {parameter:<8.6f}")
        return formatted_params

    @staticmethod
    def export_to_param(formatted_params: list[str], filename_out: str) -> None:
        """
        Exports a list of formatted parameters to an ArduPilot parameter file.

        Args:
            formatted_params (List[str]): The list of formatted parameters to export.
            filename_out (str): The output filename.

        Returns:
            None

        """
        if not formatted_params:
            return
        try:
            # Ensure newline character is LF, even on windows
            with open(filename_out, "w", encoding="utf-8", newline="\n") as output_file:
                for line in formatted_params:
                    output_file.write(line + "\n")
        except OSError as e:
            msg = f"ERROR: writing to file {filename_out}: {e}"
            raise SystemExit(msg) from e

    @staticmethod
    def print_out(formatted_params: list[str], name: str) -> None:
        """
        Print out the contents of the provided list.

        If the list is too large, print only the ones that fit on the screen and
        wait for user input to continue.

        Args:
            formatted_params (List[str]): The list of formatted parameters to print.
            name (str): A descriptive string for the list contents

        Returns:
            None

        """
        if not formatted_params:
            return

        rows_str = "100"  # number of lines to display before waiting for user input

        # Get the size of the terminal
        if __name__ == "__main__":
            rows_str, _columns = os_popen("stty size", "r").read().split()  # noqa: S605, S607

        # Convert rows to integer
        rows = int(rows_str) - 2  # -2 for the next print and the input line

        # Convert rows
        print(f"\n{name} has {len(formatted_params)} parameters:")  # noqa: T201
        for i, line in enumerate(formatted_params):
            if i % rows == 0 and __name__ == "__main__":
                input(f"\n{name} list is long hit enter to continue")
                rows_str, _columns = os_popen("stty size", "r").read().split()  # noqa: S605, S607
                rows = int(rows_str) - 2  # -2 for the next print and the input line
            print(line)  # noqa: T201


def get_xml_data(base_url: str, directory: str, filename: str, vehicle_type: str) -> ET.Element:  # pylint: disable=too-many-locals
    """
    Fetches XML data from a local file or a URL.

    Args:
        base_url (str): The base URL for fetching the XML file.
        directory (str): The directory where the XML file is expected.
        filename (str): The name of the XML file.
        vehicle_type (str): The type of the vehicle.

    Returns:
        ET.Element: The root element of the parsed XML data.

    """
    file_path = os_path.join(directory, filename)
    # Check if the locally cached file exists
    if os_path.isfile(file_path):
        # Load the file content relative to the script location
        with open(file_path, encoding="utf-8") as file:
            xml_data = file.read()
    elif os_path.isfile(filename):
        # Load the file content from the current directory
        with open(filename, encoding="utf-8") as file:
            xml_data = file.read()
    else:
        # No locally cached file exists, get it from the internet
        try:
            # pylint: disable=import-outside-toplevel
            from requests import exceptions as requests_exceptions  # type: ignore[import-untyped]
            from requests import get as requests_get

            # pylint: enable=import-outside-toplevel
        except ImportError as exc:
            logging.critical("The requests package was not found")
            logging.critical("Please install it by running 'pip install requests' in your terminal.")
            msg = "requests package is not installed"
            raise SystemExit(msg) from exc
        # Send a GET request to the URL
        url = base_url + filename
        proxies = get_env_proxies()
        try:
            response = requests_get(url, timeout=5, proxies=proxies) if proxies else requests_get(url, timeout=5)
            if response.status_code != 200:
                logging.warning("Remote URL: %s", url)
                msg = f"HTTP status code {response.status_code}"
                raise requests_exceptions.RequestException(msg)
        except requests_exceptions.RequestException as e:
            logging.warning("Unable to fetch XML data: %s", e)
            # Send a GET request to the URL to the fallback (DEV) URL
            try:
                url = BASE_URL + vehicle_type + "/" + PARAM_DEFINITION_XML_FILE
                logging.warning("Falling back to the DEV XML file: %s", url)
                response = requests_get(url, timeout=5, proxies=proxies)
                if response.status_code != 200:
                    logging.critical("Remote URL: %s", url)
                    msg = f"HTTP status code {response.status_code}"
                    raise requests_exceptions.RequestException(msg)
            except requests_exceptions.RequestException as exp:
                logging.critical("Unable to fetch XML data: %s", exp)
                msg = "Unable to fetch online XML documentation."
                msg += f"\nDownload it manually from {url} and"
                msg += f"\nplace it in the {directory} directory"
                raise SystemExit(msg) from exp
        # Get the text content of the response
        xml_data = response.text
        try:
            # Write the content to a file
            with open(os_path.join(directory, filename), "w", encoding="utf-8") as file:
                file.write(xml_data)
        except PermissionError as e:
            logging.critical("Permission denied to write XML data to file: %s", e)
            msg = "permission denied to write online XML documentation to file"
            raise SystemExit(msg) from e

    # Parse the XML data
    return DET.fromstring(xml_data)  # type: ignore[no-any-return]


def get_env_proxies() -> Union[dict[str, str], None]:
    proxies_env = {
        "http": os_environ.get("HTTP_PROXY") or os_environ.get("http_proxy"),
        "https": os_environ.get("HTTPS_PROXY") or os_environ.get("https_proxy"),
        "no_proxy": os_environ.get("NO_PROXY") or os_environ.get("no_proxy"),
    }
    # Remove None values
    proxies_dict: dict[str, str] = {k: v for k, v in proxies_env.items() if v is not None}
    # define as None if no proxies are defined in the OS environment variables
    proxies = proxies_dict if proxies_dict else None
    if proxies:
        logging.info("Proxies: %s", proxies)
    else:
        logging.debug("Proxies: %s", proxies)
    return proxies


def load_default_param_file(directory: str) -> dict[str, "Par"]:
    param_default_dict: dict[str, Par] = {}
    # Load parameter default values if the 00_default.param file exists
    try:
        param_default_dict = Par.load_param_file_into_dict(os_path.join(directory, "00_default.param"))
    except FileNotFoundError:
        logging.warning("Default parameter file 00_default.param not found. No default values will be annotated.")
        logging.warning("Create one by using the command ./extract_param_defaults.py log_file.bin > 00_default.param")
    return param_default_dict


def remove_prefix(text: str, prefix: str) -> str:
    """
    Removes a prefix from a string.

    Args:
        text (str): The original string.
        prefix (str): The prefix to remove.

    Returns:
        str: The string without the prefix.

    """
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def split_into_lines(string_to_split: str, maximum_line_length: int) -> list[str]:
    """
    Splits a string into lines of a maximum length.

    Args:
        string_to_split (str): The string to split.
        maximum_line_length (int): The maximum length of a line.

    Returns:
        List[str]: The list of lines.

    """
    doc_lines = re.findall(r".{1," + str(maximum_line_length) + r"}(?:\s|$)", string_to_split)
    # Remove trailing whitespace from each line
    return [line.rstrip() for line in doc_lines]


def create_doc_dict(root: ET.Element, vehicle_type: str, max_line_length: int = 100) -> dict[str, Any]:
    """
    Create a dictionary of parameter documentation from the root element of the parsed XML data.

    Args:
        root (ET.Element): The root element of the parsed XML data.
        vehicle_type (str): vehicle type string.
        max_line_length (int): max line length

    Returns:
        Dict[str, Any]: A dictionary of parameter documentation.

    """
    # Dictionary to store the parameter documentation
    doc: dict[str, Any] = {}

    # Use the findall method with an XPath expression to find all "param" elements
    for param in root.findall(".//param"):
        name = param.get("name")
        if name is None:
            continue
        if vehicle_type == "Heli":
            vehicle_type = "Helicopter"
        # Remove the <vehicle_type>: prefix from the name if it exists
        name = remove_prefix(name, vehicle_type + ":")

        human_name = param.get("humanName")
        documentation = param.get("documentation")
        documentation_lst: list[str] = []
        if documentation:
            documentation_lst = split_into_lines(documentation, max_line_length)
        # the keys are the "name" attribute of the "field" sub-elements
        # the values are the text content of the "field" sub-elements
        fields = {field.get("name"): field.text for field in param.findall("field")}
        # if Units and UnitText exist, combine them into a single element
        delete_unit_text = False
        for key, value in fields.items():
            if key == "Units" and "UnitText" in fields:
                fields[key] = f"{value} ({fields['UnitText']})"
                delete_unit_text = True
        if delete_unit_text:
            del fields["UnitText"]
        # the keys are the "code" attribute of the "values/value" sub-elements
        # the values are the text content of the "values/value" sub-elements
        values = {value.get("code"): value.text for value in param.findall("values/value")}

        # Dictionary with "Parameter names" as keys and the values is a
        # dictionary with "humanName", "documentation" attributes and
        # "fields", "values" sub-elements.
        doc[name] = {
            "humanName": human_name,
            "documentation": documentation_lst,
            "fields": fields,
            "values": values,
        }

    return doc


def format_columns(values: dict[str, Any], max_width: int = 105, max_columns: int = 4) -> list[str]:
    """
    Formats a dictionary of values into column-major horizontally aligned columns.

    It uses at most max_columns columns.

    Args:
        values (Dict[str, Any]): The dictionary of values to format.
        max_width (int, optional): The maximum number of characters on all columns. Default is 105.
        max_columns (int): Maximum number of columns

    Returns:
        List[str]: The list of formatted strings.

    """
    # Convert the dictionary into a list of strings
    strings = [f"{k}: {v}" for k, v in values.items()]

    if (not strings) or (len(strings) == 0):
        return []

    # Calculate the maximum length of the strings
    max_len = max(len(s) for s in strings)

    num_cols = 1  # At least one column, no matter what max_columns is
    # Determine the number of columns
    # Column distribution will only happen if it results in more than 5 rows
    # The strings will be distributed evenly across up-to max_columns columns.
    for num_cols in range(max_columns, 0, -1):
        if len(strings) // num_cols > 5 and (max_len + 2) * num_cols < max_width:
            break

    # Calculate the column width
    col_width = max_width // num_cols

    num_rows = (len(strings) + num_cols - 1) // num_cols

    formatted_strings = []
    for j in range(num_rows):
        row = []
        for i in range(num_cols):
            if i * num_rows + j < len(strings):
                if i < num_cols - 1 and ((i + 1) * num_rows + j < len(strings)):
                    row.append(strings[i * num_rows + j].ljust(col_width))
                else:
                    row.append(strings[i * num_rows + j])
        formatted_strings.append(" ".join(row))

    return formatted_strings


def extract_parameter_name(item: str) -> str:
    """Extract the parameter name from a line. Very simple to use in sorting."""
    item = item.strip()
    match = re.match(PARAM_NAME_REGEX, item)
    return match.group(0) if match else item


def missionplanner_sort(item: str) -> tuple[str, ...]:
    """MissionPlanner parameter sorting function."""
    # Split the parameter name by underscore
    parts = extract_parameter_name(item).split("_")
    # Compare the parts separately
    return tuple(parts)


def extract_parameter_name_and_validate(line: str, filename: str, line_nr: int) -> str:
    """
    Extracts the parameter name from a line and validates it.

    Args:
        line (str): The line to extract the parameter name from.
        filename (str): filename.
        line_nr (int): line number.

    Returns:
        str: The extracted parameter name.

    Raises:
        SystemExit: If the line is invalid or the parameter name is too long or invalid.

    """
    # Extract the parameter name
    match = re.match(PARAM_NAME_REGEX, line)
    if match:
        param_name = match.group(0)
    else:
        logging.critical("Invalid line %d in file %s: %s", line_nr, filename, line)
        msg = "Invalid line in input file"
        raise SystemExit(msg)
    param_len = len(param_name)
    param_sep = line[param_len]  # the character following the parameter name must be a separator
    if param_sep not in {",", " ", "\t"}:
        logging.critical("Invalid parameter name %s on line %d in file %s", param_name, line_nr, filename)
        msg = "Invalid parameter name"
        raise SystemExit(msg)
    if param_len > PARAM_NAME_MAX_LEN:
        logging.critical("Too long parameter name on line %d in file %s", line_nr, filename)
        msg = "Too long parameter name"
        raise SystemExit(msg)
    return param_name


def update_parameter_documentation(
    doc: dict[str, Any],
    target: str = ".",
    sort_type: str = "none",
    param_default_dict: Optional[dict] = None,
    delete_documentation_annotations: bool = False,
) -> None:
    """
    Updates the parameter documentation in the target file or in all *.param,*.parm files of the target directory.

    This function iterates over all the ArduPilot parameter files in the target directory or file.
    For each file, it DELETES all comments that start at the beginning of a line, optionally sorts the
    parameter names and checks if the parameter name is in the dictionary of parameter documentation.
    If it is, it prefixes the line with a comment derived from the dictionary element.
    If it's not, it copies the parameter line 1-to-1.
    After processing all the parameters in a file, it writes the new lines back to the file.

    Args:
        doc (Dict[str, Any]): A dictionary of parameter documentation.
        target (str, optional): The target directory or file. Default is '.'.
        sort_type (str, optional): The type of sorting to apply to the parameters.
                                   Can be 'none', 'missionplanner', or 'mavproxy'. Default is 'none'.
        param_default_dict (Dict, optional): A dictionary of default parameter values. Default is None.
                                              If None, an empty dictionary is used.
        delete_documentation_annotations (bool): delete documentation annotations from file.

    """
    # Check if the target is a file or a directory
    if os_path.isfile(target):
        # If it's a file, process only that file
        param_files = [target]
    elif os_path.isdir(target):
        # If it's a directory, process all .param and .parm files in that directory
        param_files = glob.glob(os_path.join(target, "*.param")) + glob.glob(os_path.join(target, "*.parm"))
    else:
        msg = f"Target '{target}' is neither a file nor a directory."
        raise ValueError(msg)

    if param_default_dict is None:
        param_default_dict = {}

    # Iterate over all the target ArduPilot parameter files
    for param_file in param_files:
        if os_path.basename(param_file).endswith("24_inflight_magnetometer_fit_setup.param") and "MAGH_ALT_DELTA" not in doc:
            continue

        # Read the entire file contents
        with open(param_file, encoding="utf-8") as file:
            lines = file.readlines()

        update_parameter_documentation_file(
            doc, sort_type, param_default_dict, param_file, lines, delete_documentation_annotations
        )


def update_parameter_documentation_file(  # pylint: disable=too-many-locals, too-many-arguments, too-many-positional-arguments
    doc: dict,
    sort_type: str,
    param_default_dict: dict,
    param_file: str,
    lines: list[str],
    delete_documentation_annotations: bool,
) -> None:
    new_lines = []

    total_params = 0
    documented_params = 0
    undocumented_params = []
    is_first_param_in_file = True
    if sort_type == "missionplanner":
        lines.sort(key=missionplanner_sort)
    if sort_type == "mavproxy":
        lines.sort(key=extract_parameter_name)
    for n, f_line in enumerate(lines, start=1):
        line = f_line.strip()
        if not line.startswith("#") and line:
            param_name = extract_parameter_name_and_validate(line, param_file, n)

            if param_name in doc and not delete_documentation_annotations:
                # If the parameter name is in the dictionary,
                #  prefix the line with a comment derived from the dictionary element
                data = doc[param_name]
                prefix_parts = [
                    f"{data['humanName']}",
                ]
                prefix_parts += data["documentation"]
                for key, value in data["fields"].items():
                    prefix_parts.append(f"{key}: {value}")
                prefix_parts += format_columns(data["values"])
                doc_text = "\n# ".join(prefix_parts)
                if param_name in param_default_dict:
                    default_value = format(param_default_dict[param_name].value, ".6f").rstrip("0").rstrip(".")
                    doc_text += f"\n# Default: {default_value}"
                if not is_first_param_in_file:
                    new_lines.append("\n")
                new_lines.append(f"# {doc_text}\n{line}\n")
                documented_params += 1
            else:
                # If the parameter name is in not the dictionary, copy the parameter line 1-to-1
                new_lines.append(f"{line}\n")
                undocumented_params.append(param_name)
            total_params += 1
            is_first_param_in_file = False

    if total_params == documented_params:
        logging.info("Read file %s with %d parameters, all got documented", param_file, total_params)
    else:
        logging.warning(
            "Read file %s with %d parameters, but only %s of which got documented", param_file, total_params, documented_params
        )
        logging.warning("No documentation found for: %s", ", ".join(undocumented_params))

    # Write the new file contents to the file
    with open(param_file, "w", encoding="utf-8", newline="\n") as file:  # Ensure newline character is LF, even on windows
        file.writelines(new_lines)


def print_read_only_params(doc: dict) -> None:
    """
    Print the names of read-only parameters.

    Args:
        doc (dict): A dictionary of parameter documentation.

    """
    logging.info("ReadOnly parameters:")
    for param_name, param_value in doc.items():
        if "ReadOnly" in param_value["fields"] and param_value["fields"]["ReadOnly"]:
            logging.info(param_name)


def get_xml_dir(target: str) -> str:
    return target if os_path.isdir(target) else os_path.dirname(os_path.realpath(target))


def get_xml_url(vehicle_type: str, firmware_version: str) -> str:
    vehicle_parm_subdir = {
        "ArduCopter": "versioned/Copter/stable-",
        "ArduPlane": "versioned/Plane/stable-",
        "Rover": "versioned/Rover/stable-",
        "ArduSub": "versioned/Sub/stable-",
        "AntennaTracker": "versioned/Tracker/stable-",
        # Not yet versioned in the https://autotest.ardupilot.org/Parameters server
        "AP_Periph": "versioned/Periph/stable-",
        "Blimp": "versioned/Blimp/stable-",
        "Heli": "versioned/Copter/stable-",
        "SITL": "versioned/SITL/stable-",
    }
    try:
        vehicle_subdir = vehicle_parm_subdir[vehicle_type] + firmware_version
    except KeyError as e:
        msg = f"Vehicle type '{vehicle_type}' is not supported."
        raise ValueError(msg) from e

    xml_url = BASE_URL
    xml_url += vehicle_subdir if firmware_version else vehicle_type
    xml_url += "/"
    return xml_url


def parse_parameter_metadata(
    xml_url: str, xml_dir: str, xml_file: str, vehicle_type: str, max_line_length: int
) -> dict[str, Any]:
    xml_root = get_xml_data(xml_url, xml_dir, xml_file, vehicle_type)
    return create_doc_dict(xml_root, vehicle_type, max_line_length)


def main() -> None:
    args = parse_arguments()
    try:
        xml_url = get_xml_url(args.vehicle_type, args.firmware_version)
        xml_dir = get_xml_dir(args.target)

        doc_dict = parse_parameter_metadata(
            xml_url, xml_dir, PARAM_DEFINITION_XML_FILE, args.vehicle_type, args.max_line_length
        )
        param_default_dict = load_default_param_file(xml_dir)
        update_parameter_documentation(
            doc_dict, args.target, args.sort, param_default_dict, args.delete_documentation_annotations
        )
        if args.verbose:
            print_read_only_params(doc_dict)

        # Annotate lua MAGfit XML documentation into the respective parameter file
        xml_file = LUA_PARAM_DEFINITION_XML_FILE
        target = os_path.join(os_path.dirname(args.target), "24_inflight_magnetometer_fit_setup.param")
        if os_path.isfile(os_path.join(os_path.dirname(args.target), xml_file)):
            doc_dict = parse_parameter_metadata(xml_url, xml_dir, xml_file, args.vehicle_type, args.max_line_length)
            param_default_dict = load_default_param_file(xml_dir)
            update_parameter_documentation(
                doc_dict, target, args.sort, param_default_dict, args.delete_documentation_annotations
            )
        else:
            logging.warning("No LUA MAGfit XML documentation found, skipping annotation of %s", target)

    except (OSError, SystemExit) as exp:
        logging.fatal(exp)
        sys_exit(1)


if __name__ == "__main__":
    main()



"""
TKinter base classes reused in multiple parts of the code.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

# https://wiki.tcl-lang.org/page/Changing+Widget+Colors

from logging import error as logging_error
from logging import warning as logging_warning
from tkinter import ttk
from typing import Union

from ardupilot_methodic_configurator import _
from ardupilot_methodic_configurator.frontend_tkinter_show import show_tooltip


def update_combobox_width(combobox: ttk.Combobox) -> None:
    # Calculate the maximum width needed for the content
    values = combobox["values"]
    max_width = max(len(value) for value in values) if values else 0
    # Set a minimum width for the combobox
    min_width = 4  # Adjust this value as needed
    # Set the width of the combobox to the maximum width, but not less than the minimum width
    combobox.config(width=max(min_width, max_width))


class AutoResizeCombobox(ttk.Combobox):  # pylint: disable=too-many-ancestors
    """
    A custom Combobox widget that automatically resizes based on its content.

    This class extends the ttk.Combobox widget to include functionality for
    automatically adjusting its width based on the longest entry in its list of
    values. It also supports displaying a tooltip when hovering over the widget.

    Attributes:
        master: The parent container in which the Combobox is placed.
        values: A tuple of strings representing the entries in the Combobox.
        selected_element: The initially selected element in the Combobox.
        tooltip: A string representing the tooltip text to display when hovering over the widget.

    """

    def __init__(
        self, master: ttk.Frame, values: list[str], selected_element: str, tooltip: Union[None, str], *args, **kwargs
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.set_entries_tuple(values, selected_element, tooltip)

    def set_entries_tuple(self, values: list[str], selected_element: str, tooltip: Union[None, str] = None) -> None:
        self["values"] = tuple(values)
        if selected_element:
            if selected_element in values:
                self.set(selected_element)
            else:
                logging_error(_("param_file combobox selected string '%s' not in list %s"), selected_element, values)
        elif values:
            logging_warning(_("No param_file combobox element selected"))
        if values:
            update_combobox_width(self)
        if tooltip:
            show_tooltip(self, tooltip)



"""
Parameter editor table GUI.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import tkinter as tk

# from logging import warning as logging_warning
# from logging import error as logging_error
from logging import critical as logging_critical
from logging import debug as logging_debug
from logging import info as logging_info
from math import nan
from platform import system as platform_system
from sys import exit as sys_exit
from tkinter import messagebox, ttk
from typing import Any, Union

from ardupilot_methodic_configurator import _
from ardupilot_methodic_configurator.annotate_params import Par
from ardupilot_methodic_configurator.backend_filesystem import LocalFilesystem, is_within_tolerance

# from ardupilot_methodic_configurator.backend_flightcontroller import FlightController
# from ardupilot_methodic_configurator.frontend_tkinter_auto_resize_combobox import AutoResizeCombobox
from ardupilot_methodic_configurator.frontend_tkinter_base_window import BaseWindow
from ardupilot_methodic_configurator.frontend_tkinter_entry_dynamic import EntryWithDynamicalyFilteredListbox
from ardupilot_methodic_configurator.frontend_tkinter_pair_tuple_combobox import PairTupleCombobox
from ardupilot_methodic_configurator.frontend_tkinter_rich_text import get_widget_font_family_and_size
from ardupilot_methodic_configurator.frontend_tkinter_scroll_frame import ScrollFrame
from ardupilot_methodic_configurator.frontend_tkinter_show import show_tooltip

NEW_VALUE_WIDGET_WIDTH = 9


class ParameterEditorTable(ScrollFrame):  # pylint: disable=too-many-ancestors
    """
    A class to manage and display the parameter editor table within the GUI.

    This class inherits from ScrollFrame and is responsible for creating,
    managing, and updating the table that displays parameters for editing.
    """

    def __init__(self, master, local_filesystem: LocalFilesystem, parameter_editor) -> None:  # noqa: ANN001
        super().__init__(master)
        self.root = master
        self.local_filesystem = local_filesystem
        self.parameter_editor = parameter_editor
        self.current_file = ""
        self.upload_checkbutton_var: dict[str, tk.BooleanVar] = {}
        self.at_least_one_param_edited = False

        style = ttk.Style()
        style.configure("narrow.TButton", padding=0, width=4, border=(0, 0, 0, 0))

        # Prepare a dictionary that maps variable names to their values
        # These variables are used by the forced_parameters and derived_parameters in configuration_steps_*.json files
        self.variables = local_filesystem.get_eval_variables()

    def repopulate(self, selected_file: str, fc_parameters: dict[str, float], show_only_differences: bool) -> None:
        for widget in self.view_port.winfo_children():
            widget.destroy()
        self.current_file = selected_file

        # Create labels for table headers
        headers = (_("-/+"), _("Parameter"), _("Current Value"), _("New Value"), _("Unit"), _("Upload"), _("Change Reason"))
        tooltips = (
            _("Delete or add a parameter"),
            _("Parameter name must be ^[A-Z][A-Z_0-9]* and most 16 characters long"),
            _("Current value on the flight controller "),
            _("New value from the above selected intermediate parameter file"),
            _("Parameter Unit"),
            _("When selected, upload the new value to the flight controller"),
            _("Reason why respective parameter changed.")
            + "\n\n"
            + _("Documenting change reasons is crucial because it:")
            + "\n"
            + _(" * Promotes thoughtful decisions over impulsive changes")
            + "\n"
            + _(" * Provides documentation for vehicle certification requirements")
            + "\n"
            + _(" * Enables validation or suggestions from team members or AI tools")
            + "\n"
            + _(" * Preserves your reasoning for future reference or troubleshooting"),
        )

        for i, header in enumerate(headers):
            label = ttk.Label(self.view_port, text=header)
            label.grid(row=0, column=i, sticky="ew")  # Use sticky="ew" to make the label stretch horizontally
            show_tooltip(label, tooltips[i])

        self.upload_checkbutton_var = {}

        # re-compute derived parameters because the fc_parameters values might have changed
        if self.local_filesystem.configuration_steps and selected_file in self.local_filesystem.configuration_steps:
            self.variables["fc_parameters"] = fc_parameters
            error_msg = self.local_filesystem.compute_parameters(
                selected_file, self.local_filesystem.configuration_steps[selected_file], "derived", self.variables
            )
            if error_msg:
                messagebox.showerror(_("Error in derived parameters"), error_msg)
            # merge derived parameter values
            elif self.local_filesystem.merge_forced_or_derived_parameters(
                selected_file, self.local_filesystem.derived_parameters, list(fc_parameters.keys())
            ):
                self.at_least_one_param_edited = True

            self.rename_fc_connection(selected_file)

        if show_only_differences:
            # recompute different_params because of renames and derived values changes
            different_params = {
                param_name: file_value
                for param_name, file_value in self.local_filesystem.file_parameters[selected_file].items()
                if param_name not in fc_parameters
                or (
                    param_name in fc_parameters and not is_within_tolerance(fc_parameters[param_name], float(file_value.value))
                )
            }
            self.__update_table(different_params, fc_parameters)
            if not different_params:
                info_msg = _("No different parameters found in {selected_file}. Skipping...").format(**locals())
                logging_info(info_msg)
                messagebox.showinfo(_("ArduPilot methodic configurator"), info_msg)
                self.parameter_editor.on_skip_click(force_focus_out_event=False)
                return
        else:
            self.__update_table(self.local_filesystem.file_parameters[selected_file], fc_parameters)
        # Scroll to the top of the parameter table
        self.canvas.yview("moveto", 0)

    def rename_fc_connection(self, selected_file: str) -> None:
        renames = {}
        if "rename_connection" in self.local_filesystem.configuration_steps[selected_file]:
            new_connection_prefix = self.local_filesystem.configuration_steps[selected_file]["rename_connection"]
            new_connection_prefix = eval(str(new_connection_prefix), {}, self.variables)  # noqa: S307 pylint: disable=eval-used
            for param_name in self.local_filesystem.file_parameters[selected_file]:
                new_prefix = new_connection_prefix
                old_prefix = param_name.split("_")[0]

                # Handle CAN parameter names peculiarities
                if new_connection_prefix[:-1] == "CAN" and "CAN_P" in param_name:
                    old_prefix = param_name.split("_")[0] + "_" + param_name.split("_")[1]
                    new_prefix = "CAN_P" + new_connection_prefix[-1]
                if new_connection_prefix[:-1] == "CAN" and "CAN_D" in param_name:
                    old_prefix = param_name.split("_")[0] + "_" + param_name.split("_")[1]
                    new_prefix = "CAN_D" + new_connection_prefix[-1]

                if new_connection_prefix[:-1] in old_prefix:
                    renames[param_name] = param_name.replace(old_prefix, new_prefix)

        new_names = set()
        for old_name, new_name in renames.items():
            if new_name in new_names:
                self.local_filesystem.file_parameters[selected_file].pop(old_name)
                logging_info(_("Removing duplicate parameter %s"), old_name)
                info_msg = _("The parameter '{old_name}' was removed due to duplication.")
                messagebox.showinfo(_("Parameter Removed"), info_msg.format(**locals()))
                # will ask the user to save changes before switching to another file
                self.at_least_one_param_edited = True
            else:
                new_names.add(new_name)
                if new_name != old_name:
                    self.local_filesystem.file_parameters[selected_file][new_name] = self.local_filesystem.file_parameters[
                        selected_file
                    ].pop(old_name)
                    logging_info(_("Renaming parameter %s to %s"), old_name, new_name)
                    info_msg = _(
                        "The parameter '{old_name}' was renamed to '{new_name}'.\n"
                        "to obey the flight controller connection defined in the component editor window."
                    )
                    messagebox.showinfo(_("Parameter Renamed"), info_msg.format(**locals()))
                    # will ask the user to save changes before switching to another file
                    self.at_least_one_param_edited = True

    def __update_table(self, params: dict[str, Par], fc_parameters: dict[str, float]) -> None:
        current_param_name: str = ""
        try:
            for i, (param_name, param) in enumerate(params.items(), 1):
                current_param_name = param_name
                param_metadata = self.local_filesystem.doc_dict.get(param_name, {})
                param_default = self.local_filesystem.param_default_dict.get(param_name, None)
                doc_tooltip = param_metadata.get(
                    "doc_tooltip", _("No documentation available in apm.pdef.xml for this parameter")
                )

                column: list[tk.Widget] = []
                column.append(self.__create_delete_button(param_name))
                column.append(self.__create_parameter_name(param_name, param_metadata, doc_tooltip))
                column.append(self.__create_flightcontroller_value(fc_parameters, param_name, param_default, doc_tooltip))
                column.append(self.__create_new_value_entry(param_name, param, param_metadata, param_default, doc_tooltip))
                column.append(self.__create_unit_label(param_metadata))
                column.append(self.__create_upload_checkbutton(param_name, bool(fc_parameters)))
                # workaround a mypy issue
                column.append(self.__create_change_reason_entry(param_name, param, column[3]))  # type: ignore[arg-type]

                column[0].grid(row=i, column=0, sticky="w", padx=0)
                column[1].grid(row=i, column=1, sticky="w", padx=0)
                column[2].grid(row=i, column=2, sticky="e", padx=0)
                column[3].grid(row=i, column=3, sticky="e", padx=0)
                column[4].grid(row=i, column=4, sticky="e", padx=0)
                column[5].grid(row=i, column=5, sticky="e", padx=0)
                column[6].grid(row=i, column=6, sticky="ew", padx=(0, 5))

            # Add the "Add" button at the bottom of the table
            add_button = ttk.Button(
                self.view_port, text=_("Add"), style="narrow.TButton", command=lambda: self.__on_parameter_add(fc_parameters)
            )
            tooltip_msg = _("Add a parameter to the {self.current_file} file")
            show_tooltip(add_button, tooltip_msg.format(**locals()))
            add_button.grid(row=len(params) + 2, column=0, sticky="w", padx=0)

        except KeyError as e:
            logging_critical(
                _("Parameter %s not found in the %s file: %s"), current_param_name, self.current_file, e, exc_info=True
            )
            sys_exit(1)

        # Configure the table_frame to stretch columns
        self.view_port.columnconfigure(0, weight=0)  # Delete and Add buttons
        self.view_port.columnconfigure(1, weight=0, minsize=120)  # Parameter name
        self.view_port.columnconfigure(2, weight=0)  # Current Value
        self.view_port.columnconfigure(3, weight=0)  # New Value
        self.view_port.columnconfigure(4, weight=0)  # Units
        self.view_port.columnconfigure(5, weight=0)  # Upload to FC
        self.view_port.columnconfigure(6, weight=1)  # Change Reason

    def __create_delete_button(self, param_name: str) -> ttk.Button:
        delete_button = ttk.Button(
            self.view_port, text=_("Del"), style="narrow.TButton", command=lambda: self.__on_parameter_delete(param_name)
        )
        tooltip_msg = _("Delete {param_name} from the {self.current_file} file")
        show_tooltip(delete_button, tooltip_msg.format(**locals()))
        return delete_button

    def __create_parameter_name(self, param_name: str, param_metadata: dict[str, Any], doc_tooltip: str) -> ttk.Label:
        is_calibration = param_metadata.get("Calibration", False)
        is_readonly = param_metadata.get("ReadOnly", False)
        parameter_label = ttk.Label(
            self.view_port,
            text=param_name + (" " * (16 - len(param_name))),
            background="red"
            if is_readonly
            else "yellow"
            if is_calibration
            else ttk.Style(self.root).lookup("TFrame", "background"),
        )
        if doc_tooltip:
            show_tooltip(parameter_label, doc_tooltip)
        return parameter_label

    def __create_flightcontroller_value(
        self, fc_parameters: dict[str, float], param_name: str, param_default: Union[None, Par], doc_tooltip: str
    ) -> ttk.Label:
        if param_name in fc_parameters:
            value_str = format(fc_parameters[param_name], ".6f").rstrip("0").rstrip(".")
            if param_default is not None and is_within_tolerance(fc_parameters[param_name], param_default.value):
                # If it matches, set the background color to light blue
                flightcontroller_value = ttk.Label(self.view_port, text=value_str, background="light blue")
            else:
                # Otherwise, set the background color to the default color
                flightcontroller_value = ttk.Label(self.view_port, text=value_str)
        else:
            flightcontroller_value = ttk.Label(self.view_port, text=_("N/A"), background="orange")
        if doc_tooltip:
            show_tooltip(flightcontroller_value, doc_tooltip)
        return flightcontroller_value

    def __update_combobox_style_on_selection(
        self, combobox_widget: PairTupleCombobox, param_default: Union[None, Par], event: tk.Event
    ) -> None:
        try:
            # we want None to raise an exception
            current_value = float(combobox_widget.get_selected_key())  # type: ignore[arg-type]
            has_default_value = param_default is not None and is_within_tolerance(current_value, param_default.value)
            combobox_widget.configure(style="default_v.TCombobox" if has_default_value else "readonly.TCombobox")
            event.width = NEW_VALUE_WIDGET_WIDTH
            combobox_widget.on_combo_configure(event)
        except ValueError:
            msg = _("Could not solve the selected {combobox_widget} key to a float value.")
            logging_info(msg.format(**locals()))

    @staticmethod
    def __update_new_value_entry_text(new_value_entry: ttk.Entry, value: float, param_default: Union[None, Par]) -> None:
        if isinstance(new_value_entry, PairTupleCombobox):
            return
        new_value_entry.delete(0, tk.END)
        value_str = format(value, ".6f").rstrip("0").rstrip(".")
        new_value_entry.insert(0, value_str)
        if param_default is not None and is_within_tolerance(value, param_default.value):
            new_value_entry.configure(style="default_v.TEntry")
        else:
            new_value_entry.configure(style="TEntry")

    def __create_new_value_entry(  # pylint: disable=too-many-arguments, too-many-positional-arguments, too-many-locals
        self,
        param_name: str,
        param: Par,
        param_metadata: dict[str, Any],
        param_default: Union[None, Par],
        doc_tooltip: str,
    ) -> Union[PairTupleCombobox, ttk.Entry]:
        is_bitmask = param_metadata and "Bitmask" in param_metadata
        present_as_forced = False
        present_as_derived = False
        if (
            self.current_file in self.local_filesystem.forced_parameters
            and param_name in self.local_filesystem.forced_parameters[self.current_file]
        ):
            present_as_forced = True
            new_value = self.local_filesystem.forced_parameters[self.current_file][param_name].value
            if (is_bitmask and param.value != new_value) or not is_within_tolerance(param.value, new_value):
                param.value = new_value
                self.at_least_one_param_edited = True
        if (
            self.current_file in self.local_filesystem.derived_parameters
            and param_name in self.local_filesystem.derived_parameters[self.current_file]
        ):
            present_as_derived = True
            new_value = self.local_filesystem.derived_parameters[self.current_file][param_name].value
            if (is_bitmask and param.value != new_value) or not is_within_tolerance(param.value, new_value):
                param.value = new_value
                self.at_least_one_param_edited = True

        bitmask_dict = None
        value_str = format(param.value, ".6f").rstrip("0").rstrip(".")
        new_value_entry: Union[PairTupleCombobox, ttk.Entry]
        if (
            param_metadata
            and "values" in param_metadata
            and param_metadata["values"]
            and value_str in param_metadata["values"]
        ):
            selected_value = param_metadata["values"].get(value_str, None)
            has_default_value = param_default is not None and is_within_tolerance(param.value, param_default.value)
            new_value_entry = PairTupleCombobox(
                self.view_port,
                param_metadata["values"],
                value_str,
                param_name,
                style="TCombobox"
                if present_as_forced or present_as_derived
                else "default_v.TCombobox"
                if has_default_value
                else "readonly.TCombobox",
            )
            new_value_entry.set(selected_value)
            font_family, font_size = get_widget_font_family_and_size(new_value_entry)
            font_size -= 2 if platform_system() == "Windows" else 1
            new_value_entry.config(state="readonly", width=NEW_VALUE_WIDGET_WIDTH, font=(font_family, font_size))
            new_value_entry.bind(  # type: ignore[call-overload] # workaround a mypy issue
                "<<ComboboxSelected>>",
                lambda event: self.__update_combobox_style_on_selection(new_value_entry, param_default, event),
                "+",
            )
        else:
            new_value_entry = ttk.Entry(self.view_port, width=NEW_VALUE_WIDGET_WIDTH + 1, justify=tk.RIGHT)
            ParameterEditorTable.__update_new_value_entry_text(new_value_entry, param.value, param_default)
            bitmask_dict = param_metadata.get("Bitmask") if param_metadata else None
        try:
            old_value = self.local_filesystem.file_parameters[self.current_file][param_name].value
        except KeyError as e:
            logging_critical(_("Parameter %s not found in the %s file: %s"), param_name, self.current_file, e, exc_info=True)
            sys_exit(1)

        # Store error messages for forced and derived parameters
        forced_error_msg = _(
            "This parameter already has the correct value for this configuration step.\n"
            "You must not change it, as this would defeat the purpose of this configuration step.\n\n"
            "Add it to other configuration step and change it there if you have a good reason to."
        )
        derived_error_msg = _(
            "This parameter value has been derived from information you entered in the component editor window.\n"
            "You need to change the information on that window to update the value here.\n"
        )

        # Function to show the appropriate error message
        def show_parameter_error(_event: tk.Event) -> None:
            if present_as_forced:
                messagebox.showerror(_("Forced Parameter"), forced_error_msg)
            elif present_as_derived:
                messagebox.showerror(_("Derived Parameter"), derived_error_msg)

        if present_as_forced or present_as_derived:
            new_value_entry.config(state="disabled", background="light grey")
            new_value_entry.bind("<Button-1>", show_parameter_error)
            # Also bind to right-click for completeness
            new_value_entry.bind("<Button-3>", show_parameter_error)
        elif bitmask_dict:
            new_value_entry.bind(
                "<Double-Button>",
                lambda event: self.__open_bitmask_selection_window(event, param_name, bitmask_dict, old_value),
            )
            # pylint: disable=line-too-long
            new_value_entry.bind(
                "<FocusOut>",
                lambda event, current_file=self.current_file, param_name=param_name: self.__on_parameter_value_change(  # type: ignore[misc]
                    event, current_file, param_name
                ),
            )
            # pylint: enable=line-too-long
        else:
            # pylint: disable=line-too-long
            new_value_entry.bind(
                "<FocusOut>",
                lambda event, current_file=self.current_file, param_name=param_name: self.__on_parameter_value_change(  # type: ignore[misc]
                    event, current_file, param_name
                ),
            )
            # pylint: enable=line-too-long
        if doc_tooltip:
            show_tooltip(new_value_entry, doc_tooltip)
        return new_value_entry

    def __open_bitmask_selection_window(self, event: tk.Event, param_name: str, bitmask_dict: dict, old_value: float) -> None:  # pylint: disable=too-many-locals
        def on_close() -> None:
            checked_keys = [key for key, var in checkbox_vars.items() if var.get()]
            # Convert checked keys back to a decimal value
            new_decimal_value = sum(1 << key for key in checked_keys)
            # Update new_value_entry with the new decimal value
            ParameterEditorTable.__update_new_value_entry_text(
                event.widget, new_decimal_value, self.local_filesystem.param_default_dict.get(param_name, None)
            )
            self.at_least_one_param_edited = (old_value != new_decimal_value) or self.at_least_one_param_edited
            self.local_filesystem.file_parameters[self.current_file][param_name].value = new_decimal_value
            # Destroy the window
            window.destroy()
            # Issue a FocusIn event on something else than new_value_entry to prevent endless looping
            self.root.focus_set()
            # Run the Tk event loop once to process the event
            self.root.update_idletasks()
            # Re-bind the FocusIn event to new_value_entry
            event.widget.bind(
                "<Double-Button>",
                lambda event: self.__open_bitmask_selection_window(event, param_name, bitmask_dict, old_value),
            )

        def is_widget_visible(widget: Union[tk.Misc, None]) -> bool:
            return bool(widget and widget.winfo_ismapped())

        def focus_out_handler(_event: tk.Event) -> None:
            if not is_widget_visible(window.focus_get()):
                on_close()

        def get_param_value_msg(_param_name: str, checked_keys: set) -> str:
            _new_decimal_value = sum(1 << key for key in checked_keys)
            text = _("{_param_name} Value: {_new_decimal_value}")
            return text.format(**locals())

        def update_label() -> None:
            checked_keys = {key for key, var in checkbox_vars.items() if var.get()}
            close_label.config(text=get_param_value_msg(param_name, checked_keys))

        # Temporarily unbind the FocusIn event to prevent triggering the window again
        event.widget.unbind("<Double-Button>")
        window = tk.Toplevel(self.root)
        title = _("Select {param_name} Bitmask Options")
        window.title(title.format(**locals()))
        checkbox_vars = {}

        main_frame = ttk.Frame(window)
        main_frame.pack(expand=True, fill=tk.BOTH)

        # Convert current_value to a set of checked keys
        current_value = int(event.widget.get())
        checked_keys = {key for key, _value in bitmask_dict.items() if (current_value >> key) & 1}

        for i, (key, value) in enumerate(bitmask_dict.items()):
            var = tk.BooleanVar(value=key in checked_keys)
            checkbox_vars[key] = var
            checkbox = ttk.Checkbutton(main_frame, text=value, variable=var, command=update_label)
            checkbox.grid(row=i, column=0, sticky="w")

        # Replace the close button with a read-only label displaying the current new_decimal_value
        close_label = ttk.Label(main_frame, text=get_param_value_msg(param_name, checked_keys))
        close_label.grid(row=len(bitmask_dict), column=0, pady=10)

        # Bind the on_close function to the window's WM_DELETE_WINDOW protocol
        window.protocol("WM_DELETE_WINDOW", on_close)
        window.bind("<FocusOut>", focus_out_handler)
        for child in window.winfo_children():
            child.bind("<FocusOut>", focus_out_handler)

        # Make sure the window is visible before disabling the parent window
        window.deiconify()
        self.root.update_idletasks()
        window.grab_set()

        window.wait_window()  # Wait for the window to be closed

    def __create_unit_label(self, param_metadata: dict[str, Union[float, str]]) -> ttk.Label:
        unit_label = ttk.Label(self.view_port, text=param_metadata.get("unit", ""))
        unit_tooltip = str(
            param_metadata.get("unit_tooltip", _("No documentation available in apm.pdef.xml for this parameter"))
        )
        if unit_tooltip:
            show_tooltip(unit_label, unit_tooltip)
        return unit_label

    def __create_upload_checkbutton(self, param_name: str, fc_connected: bool) -> ttk.Checkbutton:
        self.upload_checkbutton_var[param_name] = tk.BooleanVar(value=fc_connected)
        upload_checkbutton = ttk.Checkbutton(self.view_port, variable=self.upload_checkbutton_var[param_name])
        upload_checkbutton.configure(state="normal" if fc_connected else "disabled")
        msg = _("When selected upload {param_name} new value to the flight controller")
        show_tooltip(upload_checkbutton, msg.format(**locals()))
        return upload_checkbutton

    def __create_change_reason_entry(
        self, param_name: str, param: Par, new_value_entry: Union[ttk.Entry, PairTupleCombobox]
    ) -> ttk.Entry:
        present_as_forced = False
        if (
            self.current_file in self.local_filesystem.forced_parameters
            and param_name in self.local_filesystem.forced_parameters[self.current_file]
        ):
            present_as_forced = True
            if param.comment != self.local_filesystem.forced_parameters[self.current_file][param_name].comment:
                param.comment = self.local_filesystem.forced_parameters[self.current_file][param_name].comment
                self.at_least_one_param_edited = True
        if (
            self.current_file in self.local_filesystem.derived_parameters
            and param_name in self.local_filesystem.derived_parameters[self.current_file]
        ):
            present_as_forced = True
            if param.comment != self.local_filesystem.derived_parameters[self.current_file][param_name].comment:
                param.comment = self.local_filesystem.derived_parameters[self.current_file][param_name].comment
                self.at_least_one_param_edited = True

        change_reason_entry = ttk.Entry(self.view_port, background="white")
        change_reason_entry.insert(0, "" if param.comment is None else param.comment)
        if present_as_forced:
            change_reason_entry.config(state="disabled", background="light grey")
        else:
            # pylint: disable=line-too-long
            change_reason_entry.bind(
                "<FocusOut>",
                lambda event, current_file=self.current_file, param_name=param_name: self.__on_parameter_change_reason_change(  # type: ignore[misc]
                    event, current_file, param_name
                ),
            )
            # pylint: enable=line-too-long
        _value = new_value_entry.get()
        msg = _("Reason why {param_name} should change to {_value}")
        show_tooltip(change_reason_entry, msg.format(**locals()))
        return change_reason_entry

    def __on_parameter_delete(self, param_name: str) -> None:
        msg = _("Are you sure you want to delete the {param_name} parameter?")
        if messagebox.askyesno(f"{self.current_file}", msg.format(**locals())):
            # Capture current vertical scroll position
            current_scroll_position = self.canvas.yview()[0]

            # Delete the parameter
            del self.local_filesystem.file_parameters[self.current_file][param_name]
            self.at_least_one_param_edited = True
            self.parameter_editor.repopulate_parameter_table(self.current_file)

            # Restore the scroll position
            self.canvas.yview_moveto(current_scroll_position)

    def __on_parameter_add(self, fc_parameters: dict[str, float]) -> None:
        add_parameter_window = BaseWindow(self.root)
        add_parameter_window.root.title(_("Add Parameter to ") + self.current_file)
        add_parameter_window.root.geometry("450x300")

        # Label for instruction
        instruction_label = ttk.Label(add_parameter_window.main_frame, text=_("Enter the parameter name to add:"))
        instruction_label.pack(pady=5)

        param_dict = self.local_filesystem.doc_dict or fc_parameters

        if not param_dict:
            messagebox.showerror(
                _("Operation not possible"),
                _("No apm.pdef.xml file and no FC connected. Not possible autocomplete parameter names."),
            )
            return

        # Remove the parameters that are already displayed in this configuration step
        possible_add_param_names = [
            param_name
            for param_name in param_dict
            if param_name not in self.local_filesystem.file_parameters[self.current_file]
        ]

        possible_add_param_names.sort()

        # Prompt the user for a parameter name
        parameter_name_combobox = EntryWithDynamicalyFilteredListbox(
            add_parameter_window.main_frame,
            possible_add_param_names,
            startswith_match=False,
            ignorecase_match=True,
            listbox_height=12,
            width=28,
        )
        parameter_name_combobox.pack(padx=5, pady=5)
        BaseWindow.center_window(add_parameter_window.root, self.root)
        parameter_name_combobox.focus()

        def custom_selection_handler(event: tk.Event) -> None:
            parameter_name_combobox.update_entry_from_listbox(event)
            if self.__confirm_parameter_addition(parameter_name_combobox.get().upper(), fc_parameters):
                add_parameter_window.root.destroy()
            else:
                add_parameter_window.root.focus()

        # Bindings to handle Enter press and selection while respecting original functionalities
        parameter_name_combobox.bind("<Return>", custom_selection_handler)
        parameter_name_combobox.bind("<<ComboboxSelected>>", custom_selection_handler)

    def __confirm_parameter_addition(self, param_name: str, fc_parameters: dict[str, float]) -> bool:
        if not param_name:
            messagebox.showerror(_("Invalid parameter name."), _("Parameter name can not be empty."))
            return False
        if param_name in self.local_filesystem.file_parameters[self.current_file]:
            messagebox.showerror(_("Invalid parameter name."), _("Parameter already exists, edit it instead"))
            return False
        if fc_parameters:
            if param_name in fc_parameters:
                self.local_filesystem.file_parameters[self.current_file][param_name] = Par(fc_parameters[param_name], "")
                self.at_least_one_param_edited = True
                self.parameter_editor.repopulate_parameter_table(self.current_file)
                return True
            messagebox.showerror(_("Invalid parameter name."), _("Parameter name not found in the flight controller."))
        elif self.local_filesystem.doc_dict:
            if param_name in self.local_filesystem.doc_dict:
                self.local_filesystem.file_parameters[self.current_file][param_name] = Par(
                    self.local_filesystem.param_default_dict.get(param_name, Par(0, "")).value, ""
                )
                self.at_least_one_param_edited = True
                self.parameter_editor.repopulate_parameter_table(self.current_file)
                return True
            error_msg = _("'{param_name}' not found in the apm.pdef.xml file.")
            messagebox.showerror(_("Invalid parameter name."), error_msg.format(**locals()))
        else:
            messagebox.showerror(
                _("Operation not possible"),
                _("Can not add parameter when no FC is connected and no apm.pdef.xml file exists."),
            )
        return False

    def __on_parameter_value_change(self, event: tk.Event, current_file: str, param_name: str) -> None:
        # Get the new value from the Entry widget
        new_value = event.widget.get_selected_key() if isinstance(event.widget, PairTupleCombobox) else event.widget.get()
        try:
            old_value = self.local_filesystem.file_parameters[current_file][param_name].value
        except KeyError as e:
            logging_critical(_("Parameter %s not found in the %s file: %s"), param_name, current_file, e, exc_info=True)
            sys_exit(1)
        valid: bool = True
        changed: bool = False
        p: float = nan
        # Check if the input is a valid float
        try:
            p = float(new_value)  # type: ignore[arg-type] # workaround a mypy bug
            changed = not is_within_tolerance(old_value, p)
            param_metadata = self.local_filesystem.doc_dict.get(param_name, {})
            p_min = param_metadata.get("min", None)
            p_max = param_metadata.get("max", None)
            if changed:
                if p_min and p < p_min:
                    msg = _("The value for {param_name} {p} should be greater than {p_min}\n")
                    if not messagebox.askyesno(
                        _("Out-of-bounds Value"), msg.format(**locals()) + _("Use out-of-bounds value?"), icon="warning"
                    ):
                        valid = False
                if p_max and p > p_max:
                    msg = _("The value for {param_name} {p} should be smaller than {p_max}\n")
                    if not messagebox.askyesno(
                        _("Out-of-bounds Value"), msg.format(**locals()) + _("Use out-of-bounds value?"), icon="warning"
                    ):
                        valid = False
        except ValueError:
            # Optionally, you can handle the invalid value here, for example, by showing an error message
            error_msg = _("The value for {param_name} must be a valid float.")
            messagebox.showerror(_("Invalid Value"), error_msg.format(**locals()))
            valid = False
        if valid:
            if changed and not self.at_least_one_param_edited:
                logging_debug(_("Parameter %s changed, will later ask if change(s) should be saved to file."), param_name)
            self.at_least_one_param_edited = changed or self.at_least_one_param_edited
            # Update the params dictionary with the new value
            self.local_filesystem.file_parameters[current_file][param_name].value = p
        else:
            # Revert to the previous (valid) value
            p = old_value
        self.__update_new_value_entry_text(event.widget, p, self.local_filesystem.param_default_dict.get(param_name, None))
        self.__update_change_reason_entry_tooltip(param_name, p)

    def __update_change_reason_entry_tooltip(self, param_name: str, param_value: float) -> None:
        """Update the tooltip on the change reason entry with the current parameter value."""
        value_str = format(param_value, ".6f").rstrip("0").rstrip(".")

        # Find the change reason entry widget for this parameter
        for widget in self.view_port.winfo_children():
            if isinstance(widget, ttk.Entry) and widget.grid_info().get("column", 0) == 6:
                # Get the parameter label in the same row
                row = widget.grid_info().get("row")
                for param_widget in self.view_port.winfo_children():
                    if (
                        isinstance(param_widget, ttk.Label)
                        and param_widget.grid_info().get("column", 0) == 1
                        and param_widget.grid_info().get("row") == row
                        and param_widget.cget("text").strip() == param_name
                    ):
                        # Found the matching change reason entry, update its tooltip
                        msg = _("Reason why {param_name} should change to {value_str}")
                        show_tooltip(widget, msg.format(**locals()))
                        return

        logging_debug(_("Could not find change reason entry widget for parameter %s (%s)"), param_name, value_str)

    def __on_parameter_change_reason_change(self, event: tk.Event, current_file: str, param_name: str) -> None:
        # Get the new value from the Entry widget
        new_value = event.widget.get()
        try:
            changed = new_value != self.local_filesystem.file_parameters[current_file][param_name].comment and not (
                new_value == "" and self.local_filesystem.file_parameters[current_file][param_name].comment is None
            )
        except KeyError as e:
            logging_critical(
                _("Parameter %s not found in the %s file %s: %s"), param_name, current_file, new_value, e, exc_info=True
            )
            sys_exit(1)
        if changed and not self.at_least_one_param_edited:
            logging_debug(
                _("Parameter %s change reason changed from %s to %s, will later ask if change(s) should be saved to file."),
                param_name,
                self.local_filesystem.file_parameters[current_file][param_name].comment,
                new_value,
            )
        self.at_least_one_param_edited = changed or self.at_least_one_param_edited
        # Update the params dictionary with the new value
        self.local_filesystem.file_parameters[current_file][param_name].comment = new_value

    def get_upload_selected_params(self, current_file: str) -> dict[str, Par]:
        selected_params = {}
        for param_name, checkbutton_state in self.upload_checkbutton_var.items():
            if checkbutton_state.get():
                selected_params[param_name] = self.local_filesystem.file_parameters[current_file][param_name]
        return selected_params

    def generate_edit_widgets_focus_out(self) -> None:
        # Trigger the <FocusOut> event for all entry widgets to ensure all changes are processed
        for widget in self.view_port.winfo_children():
            if isinstance(widget, ttk.Entry):
                widget.event_generate("<FocusOut>", when="now")

    def get_at_least_one_param_edited(self) -> bool:
        return self.at_least_one_param_edited

    def set_at_least_one_param_edited(self, value: bool) -> None:
        self.at_least_one_param_edited = value



#!/usr/bin/env python3

"""
Component editor GUI tests.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import contextlib
import tkinter as tk
from tkinter import ttk
from unittest.mock import MagicMock, patch

import pytest

from ardupilot_methodic_configurator.backend_filesystem import LocalFilesystem
from ardupilot_methodic_configurator.frontend_tkinter_component_editor_base import ComponentEditorWindowBase


@pytest.fixture
def editor_with_mocked_root() -> ComponentEditorWindowBase:
    """Create a mock ComponentEditorWindowBase for testing."""
    # Create the class without initialization
    with patch.object(ComponentEditorWindowBase, "__init__", return_value=None):
        editor = ComponentEditorWindowBase()  # pylint: disable=no-value-for-parameter

        # Set up all required attributes manually
        editor.root = MagicMock()
        editor.main_frame = MagicMock()
        editor.scroll_frame = MagicMock()
        editor.scroll_frame.view_port = MagicMock()

        # Mock filesystem and methods with proper schema loading
        editor.local_filesystem = MagicMock(spec=LocalFilesystem)
        editor.local_filesystem.vehicle_dir = "dummy_vehicle_dir"

        # Create a method that always returns a valid tuple regardless of input path
        editor.local_filesystem.get_component_property_description = MagicMock(return_value=("Test description", False))

        # Mock the vehicle_components attribute
        mock_vehicle_components = MagicMock()
        # Make schema loading return a valid empty schema
        mock_vehicle_components.load_schema.return_value = {"properties": {}}
        mock_vehicle_components.get_component_property_description = MagicMock(return_value=("Test description", False))
        editor.local_filesystem.vehicle_components = mock_vehicle_components

        # Setup test data
        editor.entry_widgets = {}
        editor.data = {"Components": {"Motor": {"Type": "brushless", "KV": 1000}}}

        # Mock the actual _add_widget method to avoid infinite recursion
        original_add_widget = editor._add_widget  # pylint: disable=protected-access

        # Store the original method
        editor._original_add_widget = original_add_widget  # pylint: disable=protected-access

        yield editor


@patch("tkinter.messagebox.askyesnocancel")
def test_on_closing_save(mock_dialog, editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test the on_closing method when user chooses to save."""
    mock_dialog.return_value = True  # User selects "Yes"

    # Create a replacement for save_component_json
    editor_with_mocked_root.save_component_json = MagicMock()

    # Patch sys_exit in the module where it's imported
    with patch("ardupilot_methodic_configurator.frontend_tkinter_component_editor_base.sys_exit") as mock_exit:
        with contextlib.suppress(SystemExit):
            editor_with_mocked_root.on_closing()

        # Verify save was called
        editor_with_mocked_root.save_component_json.assert_called_once()
        mock_exit.assert_called_once_with(0)


@patch("tkinter.messagebox.askyesnocancel")
def test_on_closing_no_save(mock_dialog, editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test the on_closing method when user chooses not to save."""
    mock_dialog.return_value = False  # User selects "No"

    # Create a replacement for save_component_json to avoid actual save logic
    editor_with_mocked_root.save_component_json = MagicMock()

    # Patch sys_exit in the module where it's imported
    with patch("ardupilot_methodic_configurator.frontend_tkinter_component_editor_base.sys_exit") as mock_exit:
        with contextlib.suppress(SystemExit):
            editor_with_mocked_root.on_closing()

        # Verify save was NOT called but window was destroyed
        editor_with_mocked_root.save_component_json.assert_not_called()
        editor_with_mocked_root.root.destroy.assert_called_once()
        mock_exit.assert_called_once_with(0)


@patch("tkinter.messagebox.askyesnocancel")
def test_on_closing_cancel(mock_dialog, editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test the on_closing method when user cancels."""
    mock_dialog.return_value = None  # User selects "Cancel"

    # Make sure save_component_json is a MagicMock for this test
    editor_with_mocked_root.save_component_json = MagicMock()

    with patch("ardupilot_methodic_configurator.frontend_tkinter_component_editor_base.sys_exit") as mock_exit:
        editor_with_mocked_root.on_closing()

        # Verify neither save nor destroy were called
        editor_with_mocked_root.save_component_json.assert_not_called()
        editor_with_mocked_root.root.destroy.assert_not_called()
        mock_exit.assert_not_called()


def test_update_json_data(editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test the update_json_data method."""
    # Test when Format version is not in data
    editor_with_mocked_root.data = {}
    editor_with_mocked_root.update_json_data()
    assert editor_with_mocked_root.data["Format version"] == 1

    # Test when Format version is already in data
    editor_with_mocked_root.data = {"Format version": 2}
    editor_with_mocked_root.update_json_data()
    assert editor_with_mocked_root.data["Format version"] == 2


def test_add_entry_or_combobox(editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test the add_entry_or_combobox method."""
    with patch("tkinter.ttk.Entry") as mock_entry:
        mock_entry_instance = MagicMock()
        mock_entry.return_value = mock_entry_instance

        entry_frame = MagicMock()
        result = editor_with_mocked_root.add_entry_or_combobox(42, entry_frame, ("Motor", "Type", "brushless"))

        mock_entry.assert_called_once_with(entry_frame)
        mock_entry_instance.insert.assert_called_once_with(0, "42")
        assert result == mock_entry_instance


def test_set_component_value_and_update_ui(editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test the _set_component_value_and_update_ui method."""
    # Setup test data
    editor_with_mocked_root.data = {"Components": {"Motor": {"Type": "old_value"}}}
    mock_entry = MagicMock()
    editor_with_mocked_root.entry_widgets = {("Motor", "Type"): mock_entry}

    # Call the method
    editor_with_mocked_root._set_component_value_and_update_ui(("Motor", "Type"), "new_value")  # pylint: disable=protected-access

    # Assert data was updated
    assert editor_with_mocked_root.data["Components"]["Motor"]["Type"] == "new_value"

    # Assert UI was updated
    mock_entry.delete.assert_called_once_with(0, tk.END)
    mock_entry.insert.assert_called_once_with(0, "new_value")
    mock_entry.config.assert_called_once_with(state="disabled")


@patch("tkinter.messagebox.askyesno")
def test_validate_and_save_component_json_yes(mock_dialog, editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test when user confirms saving component data."""
    mock_dialog.return_value = True

    # Create a replacement for save_component_json to avoid actual save logic
    editor_with_mocked_root.save_component_json = MagicMock()

    editor_with_mocked_root.validate_and_save_component_json()

    mock_dialog.assert_called_once()
    editor_with_mocked_root.save_component_json.assert_called_once()


@patch("tkinter.messagebox.askyesno")
def test_validate_and_save_component_json_no(mock_dialog, editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test when user cancels saving component data."""
    mock_dialog.return_value = False

    # Create a replacement for save_component_json to avoid actual save logic
    editor_with_mocked_root.save_component_json = MagicMock()

    editor_with_mocked_root.validate_and_save_component_json()

    mock_dialog.assert_called_once()
    editor_with_mocked_root.save_component_json.assert_not_called()


def test_save_component_json(editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test the save_component_json method with successful save."""
    # Setup test data
    editor_with_mocked_root.data = {"Components": {"Motor": {"Type": "brushless"}}}
    mock_entry = MagicMock()
    mock_entry.get.return_value = "new_value"
    editor_with_mocked_root.entry_widgets = {("Motor", "Type"): mock_entry}
    editor_with_mocked_root.local_filesystem.save_vehicle_components_json_data.return_value = (False, "")

    # For this test we need to use the real save_component_json, not a mock
    original_method = ComponentEditorWindowBase.save_component_json
    editor_with_mocked_root.save_component_json = lambda: original_method(editor_with_mocked_root)

    # Call the method
    editor_with_mocked_root.save_component_json()

    # Assert data was updated and saved
    assert editor_with_mocked_root.data["Components"]["Motor"]["Type"] == "new_value"
    editor_with_mocked_root.local_filesystem.save_vehicle_components_json_data.assert_called_once()
    editor_with_mocked_root.root.destroy.assert_called_once()


@patch("ardupilot_methodic_configurator.frontend_tkinter_component_editor_base.show_error_message")
def test_save_component_json_failure(mock_error_message, editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test the save_component_json method with failed save."""
    # Setup test data
    editor_with_mocked_root.data = {"Components": {"Motor": {"Type": "brushless"}}}
    mock_entry = MagicMock()
    mock_entry.get.return_value = "new_value"
    editor_with_mocked_root.entry_widgets = {("Motor", "Type"): mock_entry}
    editor_with_mocked_root.local_filesystem.save_vehicle_components_json_data.return_value = (True, "Error message")

    # For this test we need to use the real save_component_json, not a mock
    original_method = ComponentEditorWindowBase.save_component_json
    editor_with_mocked_root.save_component_json = lambda: original_method(editor_with_mocked_root)

    # Call the method
    editor_with_mocked_root.save_component_json()

    # Assert error message was shown
    mock_error_message.assert_called_once()
    editor_with_mocked_root.root.destroy.assert_called_once()


def test_add_argparse_arguments() -> None:
    """Test adding command line arguments."""
    parser = MagicMock()

    result = ComponentEditorWindowBase.add_argparse_arguments(parser)

    parser.add_argument.assert_called_once()
    assert result == parser


@patch("tkinter.ttk.LabelFrame")
@patch("tkinter.ttk.Frame")
@patch("tkinter.ttk.Label")
def test_add_widget_dict(mock_label, mock_frame, mock_labelframe, editor_with_mocked_root) -> None:  # noqa: ARG001 # pylint: disable=redefined-outer-name, unused-argument
    """Test the _add_widget method with dictionary values."""
    # Create a dictionary to test with
    test_dict = {"SubKey1": "Value1", "SubKey2": "Value2"}
    mock_parent = MagicMock()
    mock_labelframe_instance = MagicMock()
    mock_labelframe.return_value = mock_labelframe_instance

    # Create a spy on the _add_widget method that tracks calls without changing behavior
    original_add_widget = editor_with_mocked_root._add_widget  # pylint: disable=protected-access
    spy_add_widget = MagicMock()

    # Replace the method with a special version that calls the original but tracks recursion
    def special_add_widget(parent, key, value, path) -> None:
        # Record the call
        spy_add_widget(parent, key, value, path)

        # Stop recursion for dictionary values - don't actually process children
        if isinstance(value, dict):
            # Create the LabelFrame but don't recurse
            frame = ttk.LabelFrame(parent, text=key)
            frame.pack(fill="x", expand=True, padx=10, pady=5)
            return

        # For leaf values, just call the original to prevent infinite recursion
        if not isinstance(value, dict):
            original_add_widget(parent, key, value, path)

    # Replace with our special method
    editor_with_mocked_root._add_widget = special_add_widget  # pylint: disable=protected-access

    try:
        # Call the method with the test dictionary
        editor_with_mocked_root._add_widget(mock_parent, "TestKey", test_dict, [])  # pylint: disable=protected-access

        # Verify LabelFrame was created
        mock_labelframe.assert_called_once()
        mock_labelframe_instance.pack.assert_called_once()

        # Verify the method was called for the top level
        spy_add_widget.assert_any_call(mock_parent, "TestKey", test_dict, [])
    finally:
        # Restore the original method
        editor_with_mocked_root._add_widget = original_add_widget  # pylint: disable=protected-access


@patch("tkinter.ttk.Frame")
@patch("tkinter.ttk.Label")
def test_add_widget_leaf(mock_label, mock_frame, editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name, unused-argument # noqa: ARG001
    """Test the _add_widget method with leaf values."""
    # Create a completely separate mock for this test
    mock_parent = MagicMock()
    mock_frame_instance = MagicMock()
    mock_frame.return_value = mock_frame_instance

    # Mock add_entry_or_combobox to return a mock entry
    mock_entry = MagicMock()
    editor_with_mocked_root.add_entry_or_combobox = MagicMock(return_value=mock_entry)

    # Store original method to restore later
    original_add_leaf_widget = editor_with_mocked_root._ComponentEditorWindowBase__add_leaf_widget  # pylint: disable=protected-access

    # We need to directly call the original _add_widget method, but mock the private __add_leaf_widget
    # so it adds the entry to entry_widgets without creating UI elements that will fail in tests
    def mock_add_leaf_widget(parent, key, value, path) -> None:  # pylint: disable=unused-argument # noqa: ARG001
        # This simulates what the real method does, but in a test-friendly way
        # The real method would create a frame, but we'll use our mock_frame_instance
        entry = editor_with_mocked_root.add_entry_or_combobox(value, mock_frame_instance, (*path, key))
        editor_with_mocked_root.entry_widgets[(*path, key)] = entry

    # Replace the private method with our mock
    editor_with_mocked_root._ComponentEditorWindowBase__add_leaf_widget = mock_add_leaf_widget  # pylint: disable=protected-access

    try:
        # Call the method with a leaf value and path
        path = ["Component"]
        test_value = 42.5
        editor_with_mocked_root._add_widget(mock_parent, "TestKey", test_value, path)  # pylint: disable=protected-access

        # Verify entry was created through add_entry_or_combobox with correct parameters
        editor_with_mocked_root.add_entry_or_combobox.assert_called_once_with(
            test_value, mock_frame_instance, (*path, "TestKey")
        )

        # Verify the entry is stored in the entry_widgets dictionary
        expected_key = (*path, "TestKey")
        assert expected_key in editor_with_mocked_root.entry_widgets
    finally:
        # Restore original method
        editor_with_mocked_root._ComponentEditorWindowBase__add_leaf_widget = original_add_leaf_widget  # pylint: disable=protected-access


def test_populate_frames(editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test the populate_frames method."""
    # Setup test data
    editor_with_mocked_root.data = {"Components": {"Motor": {"Type": "brushless", "KV": 1000}}}

    # Mock the _add_widget method
    editor_with_mocked_root._add_widget = MagicMock()  # pylint: disable=protected-access

    # Call the method
    editor_with_mocked_root.populate_frames()

    # Verify _add_widget was called for each top-level component
    editor_with_mocked_root._add_widget.assert_called_with(  # pylint: disable=protected-access
        editor_with_mocked_root.scroll_frame.view_port, "Motor", {"Type": "brushless", "KV": 1000}, []
    )


@patch("tkinter.ttk.LabelFrame")
@patch("tkinter.ttk.Frame")
@patch("tkinter.ttk.Label")
def test_populate_frames_full_integration(mock_label, mock_frame, mock_labelframe, editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name, too-many-locals
    """Test the populate_frames method with full integration of leaf and non-leaf widgets."""
    # Setup test data with a nested structure to test both __add_non_leaf_widget and __add_leaf_widget
    editor_with_mocked_root.data = {
        "Components": {
            "Motor": {  # This will trigger __add_non_leaf_widget
                "Type": "brushless",  # This will trigger __add_leaf_widget
                "KV": 1000,  # This will trigger __add_leaf_widget
            },
            "Battery": {  # Another non-leaf
                "Voltage": 12.6  # Another leaf
            },
        }
    }

    # Create mock instances for UI elements
    mock_labelframe_instance = MagicMock()
    mock_frame_instance = MagicMock()
    mock_label_instance = MagicMock()

    mock_labelframe.return_value = mock_labelframe_instance
    mock_frame.return_value = mock_frame_instance
    mock_label.return_value = mock_label_instance

    # Mock add_entry_or_combobox to return a mock entry
    mock_entry = MagicMock()
    editor_with_mocked_root.add_entry_or_combobox = MagicMock(return_value=mock_entry)

    # Mock the get_component_property_description method to avoid schema lookup
    editor_with_mocked_root.local_filesystem.get_component_property_description = MagicMock(
        return_value=("Test description", False)
    )

    # We need to directly modify the _add_widget implementation for testing
    original_add_widget = editor_with_mocked_root._add_widget  # pylint: disable=protected-access
    original_add_non_leaf = editor_with_mocked_root._ComponentEditorWindowBase__add_non_leaf_widget  # pylint: disable=protected-access
    original_add_leaf = editor_with_mocked_root._ComponentEditorWindowBase__add_leaf_widget  # pylint: disable=protected-access

    # Create test-friendly versions that don't rely on tkinter widgets
    def test_add_non_leaf_widget(parent, key, value, path) -> None:
        frame = mock_labelframe_instance
        editor_with_mocked_root._add_widget_calls.append(("non_leaf", parent, key, value, path))  # pylint: disable=protected-access
        # Process children
        for child_key, child_value in value.items():
            editor_with_mocked_root._add_widget(frame, child_key, child_value, [*path, key])  # pylint: disable=protected-access

    def test_add_leaf_widget(parent, key, value, path) -> None:
        frame = mock_frame_instance
        entry = editor_with_mocked_root.add_entry_or_combobox(value, frame, (*path, key))
        editor_with_mocked_root.entry_widgets[(*path, key)] = entry
        editor_with_mocked_root._add_widget_calls.append(("leaf", parent, key, value, path))  # pylint: disable=protected-access

    # Replace methods
    editor_with_mocked_root._ComponentEditorWindowBase__add_non_leaf_widget = test_add_non_leaf_widget  # pylint: disable=protected-access
    editor_with_mocked_root._ComponentEditorWindowBase__add_leaf_widget = test_add_leaf_widget  # pylint: disable=protected-access
    editor_with_mocked_root._add_widget_calls = []  # pylint: disable=protected-access

    try:
        # Call the method
        editor_with_mocked_root.populate_frames()

        # Verify the hierarchical structure of calls
        calls = editor_with_mocked_root._add_widget_calls  # pylint: disable=protected-access

        # Check that we have the right number of calls
        # 2 non-leaf (Motor, Battery) + 3 leaf (Type, KV, Voltage) = 5 calls
        assert len(calls) == 5

        # Count the different types of calls
        non_leaf_calls = [c for c in calls if c[0] == "non_leaf"]
        leaf_calls = [c for c in calls if c[0] == "leaf"]

        # We should have 2 non-leaf calls and 3 leaf calls
        assert len(non_leaf_calls) == 2
        assert len(leaf_calls) == 3

        # Check that entries were created for leaf nodes
        assert len(editor_with_mocked_root.entry_widgets) == 3
        assert ("Motor", "Type") in editor_with_mocked_root.entry_widgets
        assert ("Motor", "KV") in editor_with_mocked_root.entry_widgets
        assert ("Battery", "Voltage") in editor_with_mocked_root.entry_widgets

        # Verify add_entry_or_combobox was called for each leaf
        assert editor_with_mocked_root.add_entry_or_combobox.call_count == 3

    finally:
        # Restore original methods
        editor_with_mocked_root._add_widget = original_add_widget  # pylint: disable=protected-access
        editor_with_mocked_root._ComponentEditorWindowBase__add_non_leaf_widget = original_add_non_leaf  # pylint: disable=protected-access
        editor_with_mocked_root._ComponentEditorWindowBase__add_leaf_widget = original_add_leaf  # pylint: disable=protected-access


def test_get_component_data_from_gui(editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test the get_component_data_from_gui method."""
    # Setup test entries
    mock_entry1 = MagicMock()
    mock_entry1.get.return_value = "brushless"

    mock_entry2 = MagicMock()
    mock_entry2.get.return_value = "1000"

    mock_entry3 = MagicMock()
    mock_entry3.get.return_value = "0.5"

    editor_with_mocked_root.entry_widgets = {
        ("Motor", "Type"): mock_entry1,
        ("Motor", "KV"): mock_entry2,
        ("Motor", "Weight", "Mass"): mock_entry3,
    }

    # Call the method
    result = editor_with_mocked_root.get_component_data_from_gui("Motor")

    # Verify the result has the correct structure and values
    assert result["Type"] == "brushless"
    assert result["KV"] == 1000  # Should be converted to int
    assert "Weight" in result
    assert result["Weight"]["Mass"] == 0.5  # Should be converted to float


def test_derive_initial_template_name(editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name
    """Test the derive_initial_template_name method."""
    # This is a simple test as the base implementation just returns an empty string
    component_data = {"Type": "brushless", "KV": 1000}
    result = editor_with_mocked_root.derive_initial_template_name(component_data)
    assert result == ""


@patch("ardupilot_methodic_configurator.frontend_tkinter_component_editor_base.ComponentTemplateManager")
def test_add_template_controls(mock_template_manager, editor_with_mocked_root) -> None:  # pylint: disable=redefined-outer-name, unused-argument # noqa: ARG001
    """Test the _add_template_controls method."""
    # Setup
    mock_parent_frame = MagicMock()
    editor_with_mocked_root.template_manager = MagicMock()

    # Call the method
    editor_with_mocked_root._add_template_controls(mock_parent_frame, "Motor")  # pylint: disable=protected-access

    # Verify the template manager was called correctly
    editor_with_mocked_root.template_manager.add_template_controls.assert_called_once_with(mock_parent_frame, "Motor")


@patch("ardupilot_methodic_configurator.frontend_tkinter_component_editor_base.logging_basicConfig")
@patch("ardupilot_methodic_configurator.frontend_tkinter_component_editor_base.LocalFilesystem")
def test_main_function(mock_filesystem, mock_logging_config) -> None:
    """Test the main function of the module."""
    # Mock the arguments
    mock_args = MagicMock()
    mock_args.loglevel = "INFO"
    mock_args.vehicle_dir = "/fake/path"
    mock_args.vehicle_type = "copter"
    mock_args.allow_editing_template_files = False
    mock_args.save_component_to_system_templates = False

    # Mock filesystem instance
    mock_filesystem_instance = MagicMock()
    mock_filesystem.return_value = mock_filesystem_instance

    # Mock the application window
    mock_app = MagicMock()
    mock_root = MagicMock()
    mock_app.root = mock_root

    with (
        patch(
            "ardupilot_methodic_configurator.frontend_tkinter_component_editor_base.argument_parser", return_value=mock_args
        ) as mock_parser,
        patch(
            "ardupilot_methodic_configurator.frontend_tkinter_component_editor_base.ComponentEditorWindowBase",
            return_value=mock_app,
        ) as mock_window,
    ):
        # Import and execute the module's main block
        import importlib.util  # pylint: disable=import-outside-toplevel
        import sys  # pylint: disable=import-outside-toplevel

        spec = importlib.util.spec_from_file_location(
            "test_module", "ardupilot_methodic_configurator/frontend_tkinter_component_editor_base.py"
        )
        module = importlib.util.module_from_spec(spec)
        sys.modules["test_module"] = module

        # Execute just the __main__ block
        exec(  # pylint: disable=exec-used # noqa: S102
            """
if __name__ == "__main__":
    args = argument_parser()
    logging_basicConfig(level=logging_getLevelName(args.loglevel), format="%(asctime)s - %(levelname)s - %(message)s")
    filesystem = LocalFilesystem(
        args.vehicle_dir, args.vehicle_type, "", args.allow_editing_template_files, args.save_component_to_system_templates
    )
    app = ComponentEditorWindowBase(__version__, filesystem)
    app.root.mainloop()
""",
            {
                "__name__": "__main__",
                "argument_parser": mock_parser,
                "logging_basicConfig": mock_logging_config,
                "logging_getLevelName": lambda x: x,
                "LocalFilesystem": mock_filesystem,
                "ComponentEditorWindowBase": mock_window,
                "__version__": "test_version",
            },
        )

        # Verify the window was created with correct parameters
        mock_filesystem.assert_called_once_with(
            mock_args.vehicle_dir,
            mock_args.vehicle_type,
            "",
            mock_args.allow_editing_template_files,
            mock_args.save_component_to_system_templates,
        )
        mock_window.assert_called_once_with("test_version", mock_filesystem_instance)
        mock_root.mainloop.assert_called_once()



#!/usr/bin/env python3

"""
Tests for the backend_flightcontroller.py file.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from unittest.mock import mock_open, patch

from ardupilot_methodic_configurator.annotate_params import Par
from ardupilot_methodic_configurator.backend_flightcontroller import FlightController


def test_add_connection() -> None:
    fc = FlightController(reboot_time=7)
    assert fc.add_connection("test_connection") is True
    assert fc.add_connection("test_connection") is False
    assert fc.add_connection("") is False


def test_discover_connections() -> None:
    fc = FlightController(reboot_time=7)
    fc.discover_connections()
    assert len(fc.get_connection_tuples()) > 0


def test_connect() -> None:
    fc = FlightController(reboot_time=7)
    result = fc.connect(device="test")
    assert result == ""


def test_disconnect() -> None:
    fc = FlightController(reboot_time=7)
    fc.connect(device="test")
    fc.disconnect()
    assert fc.master is None


@patch("builtins.open", new_callable=mock_open, read_data="param1=1\nparam2=2")
@patch(
    "ardupilot_methodic_configurator.annotate_params.Par.load_param_file_into_dict",
    side_effect=lambda x: {"param1": Par(1, x), "param2": Par(2, x)},
)
def test_download_params(mock_load_param_file_into_dict, mock_file) -> None:
    fc = FlightController(reboot_time=7)
    fc.connect(device="test")
    with patch("ardupilot_methodic_configurator.backend_flightcontroller.open", mock_file):
        params, _ = fc.download_params()
    assert isinstance(params, dict)
    assert params == {"param1": 1, "param2": 2}
    mock_load_param_file_into_dict.assert_called_once_with("params.param")


@patch("builtins.open", new_callable=mock_open, read_data="param1,1\nparam2,2")
@patch(
    "ardupilot_methodic_configurator.annotate_params.Par.load_param_file_into_dict",
    side_effect=lambda x: {"param1": Par(1, x), "param2": Par(2, x)},
)
def test_set_param(mock_load_param_file_into_dict, mock_file) -> None:
    fc = FlightController(reboot_time=7)
    fc.connect(device="test")
    fc.set_param("TEST_PARAM", 1.0)
    with patch("ardupilot_methodic_configurator.backend_flightcontroller.open", mock_file):
        params, _ = fc.download_params()
    assert params.get("TEST_PARAM") is None  # Assuming the mock environment does not actually set the parameter
    mock_load_param_file_into_dict.assert_called_once_with("params.param")


def test_reset_and_reconnect() -> None:
    fc = FlightController(reboot_time=7)
    fc.connect(device="test")
    result = fc.reset_and_reconnect()
    assert result == ""


def test_upload_file() -> None:
    fc = FlightController(reboot_time=7)
    fc.connect(device="test")
    result = fc.upload_file("local.txt", "remote.txt")
    # Assuming the mock environment always returns False for upload_file
    assert result is False


def test_get_connection_tuples() -> None:
    fc = FlightController(reboot_time=7)
    fc.add_connection("test_connection")
    connections = fc.get_connection_tuples()
    assert ("test_connection", "test_connection") in connections


@patch("builtins.open", new_callable=mock_open, read_data="param1,1\nparam2,2")
@patch(
    "ardupilot_methodic_configurator.annotate_params.Par.load_param_file_into_dict",
    side_effect=lambda x: {"param1": Par(1, x), "param2": Par(2, x)},
)
def test_set_param_and_verify(mock_load_param_file_into_dict, mock_file) -> None:
    fc = FlightController(reboot_time=7)
    fc.connect(device="test")
    fc.set_param("TEST_PARAM", 1.0)
    with patch("ardupilot_methodic_configurator.backend_flightcontroller.open", mock_file):
        params, _ = fc.download_params()
    # Assuming the mock environment does not actually set the parameter
    assert params.get("TEST_PARAM") is None
    mock_load_param_file_into_dict.assert_called_once_with("params.param")


def test_download_params_via_mavftp() -> None:
    fc = FlightController(reboot_time=7)
    fc.connect(device="test")
    params, default_params = fc.download_params_via_mavftp()
    assert isinstance(params, dict)
    assert isinstance(default_params, dict)


def test_auto_detect_serial() -> None:
    fc = FlightController(reboot_time=7)
    serial_ports = fc._FlightController__auto_detect_serial()  # pylint: disable=protected-access
    assert isinstance(serial_ports, list)


def test_list_serial_ports() -> None:
    serial_ports = FlightController._FlightController__list_serial_ports()  # pylint: disable=protected-access
    assert isinstance(serial_ports, list)


def test_list_network_ports() -> None:
    network_ports = FlightController._FlightController__list_network_ports()  # pylint: disable=protected-access
    assert isinstance(network_ports, list)
    assert "tcp:127.0.0.1:5760" in network_ports


def test_request_banner() -> None:
    fc = FlightController(reboot_time=7)
    fc.connect(device="test")
    fc._FlightController__request_banner()  # pylint: disable=protected-access
    # Since we cannot verify in the mock environment, we will just ensure no exceptions are raised


def test_receive_banner_text() -> None:
    fc = FlightController(reboot_time=7)
    fc.connect(device="test")
    banner_text = fc._FlightController__receive_banner_text()  # pylint: disable=protected-access
    assert isinstance(banner_text, list)


def test_request_message() -> None:
    fc = FlightController(reboot_time=7)
    fc.connect(device="test")
    fc._FlightController__request_message(1)  # pylint: disable=protected-access
    # Since we cannot verify in the mock environment, we will just ensure no exceptions are raised


def test_create_connection_with_retry() -> None:
    fc = FlightController(reboot_time=7)
    result = fc._FlightController__create_connection_with_retry(progress_callback=None, retries=1, timeout=1)  # pylint: disable=protected-access
    assert result == ""


def test_process_autopilot_version() -> None:
    fc = FlightController(reboot_time=7)
    fc.connect(device="test")
    banner_msgs = ["ChibiOS: 123", "ArduPilot"]
    result = fc._FlightController__process_autopilot_version(None, banner_msgs)  # pylint: disable=protected-access
    assert isinstance(result, str)



#!/usr/bin/python3

"""
Tests for the extract_missing_translations.py file.

Tests the extraction of missing translations from a .po file.

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import os
import sys
import tempfile
import unittest
from unittest.mock import MagicMock, call, patch

import pytest

# Add the parent directory to the path to import the script
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

# pylint: disable=wrong-import-position
import extract_missing_translations

# pylint: enable=wrong-import-position


class TestExtractMissingTranslations(unittest.TestCase):
    """Tests for the extract_missing_translations functionality using unittest framework."""

    def setUp(self) -> None:
        # Set up a test po file path
        self.test_po_file = os.path.join(os.path.dirname(__file__), "test_extract_missing_translations_input.po")

    def test_parse_arguments_real(self) -> None:
        # Test with actual argument parser
        with patch("sys.argv", ["extract_missing_translations.py"]):
            args = extract_missing_translations.parse_arguments()
            assert args.lang_code == "all"  # Default value
            assert args.output_file == "missing_translations"  # Default value
            assert args.max_translations == 60  # Default value

        # Test with custom arguments
        with patch(
            "sys.argv",
            ["extract_missing_translations.py", "--lang-code", "ja", "--output-file", "test_out", "--max-translations", "30"],
        ):
            args = extract_missing_translations.parse_arguments()
            assert args.lang_code == "ja"
            assert args.output_file == "test_out"
            assert args.max_translations == 30

    @patch("extract_missing_translations.argparse.ArgumentParser.parse_args")
    def test_parse_arguments_mock(self, mock_parse_args) -> None:
        # Use a proper mock object with the expected attributes
        mock_args = MagicMock()
        mock_args.lang_code = "zh_CN"
        mock_args.output_file = "missing_translations"
        mock_args.max_translations = 60
        mock_parse_args.return_value = mock_args

        args = extract_missing_translations.parse_arguments()
        assert args.lang_code == "zh_CN"
        assert args.output_file == "missing_translations"
        assert args.max_translations == 60

    @patch("extract_missing_translations.open")
    @patch("extract_missing_translations.gettext.translation")
    def test_extract_missing_translations_fixed(self, mock_translation, mock_open) -> None:
        # Mock the gettext translation
        mock_translator = MagicMock()
        # Make gettext return the same string for Test string 1, but a different string for Test string 2
        mock_translator.gettext.side_effect = lambda x: x if x == "Test string 1" else "Translated"
        mock_translation.return_value = mock_translator

        # Mock file reading
        mock_file = MagicMock()
        mock_file.__enter__.return_value.readlines.return_value = [
            'msgid "Test string 1"\n',
            'msgstr ""\n',
            'msgid "Test string 2"\n',
            'msgstr "Translated"\n',
            "\n",
        ]
        mock_open.return_value = mock_file

        result = extract_missing_translations.extract_missing_translations("ja")
        assert len(result) == 1
        assert result[0][1] == "Test string 1"

    @patch("extract_missing_translations.os.path.join")
    @patch("extract_missing_translations.open")
    @patch("extract_missing_translations.gettext.translation")
    def test_extract_missing_translations_with_real_file(self, mock_translation, mock_open, mock_path_join) -> None:
        # Set the path to our test file
        mock_path_join.return_value = self.test_po_file

        # Open the actual test file for reading its content
        with open(self.test_po_file, encoding="utf-8") as f:
            actual_lines = f.readlines()

        # Mock file reading to return actual content from test file
        mock_file = MagicMock()
        mock_file.__enter__.return_value.readlines.return_value = actual_lines
        mock_open.return_value = mock_file

        # Mock the gettext translation
        mock_translator = MagicMock()
        # Make gettext return the same string to simulate missing translations
        mock_translator.gettext.side_effect = lambda x: x
        mock_translation.return_value = mock_translator

        result = extract_missing_translations.extract_missing_translations("ja")
        # We should have at least 6 missing translations from the test file
        assert len(result) == 6

        # Normalize the strings to handle escape sequences
        def normalize_string(s: str) -> str:
            return s.replace('\\"', '"').replace("\\n", "\n")

        # Normalize the actual output for comparison
        normalized_entries = [normalize_string(entry) for _, entry in result]

        # Verify specific entries that should be identified as missing
        expected_missing = [
            "ArduPilot methodic configurator is a simple GUI with a table that lists parameters. "
            "The GUI reads intermediate parameter files from a directory and displays their parameters in a table. "
            "Each row displays the parameter name, its current value on the flight controller, its new value from "
            'the selected intermediate parameter file, and an "Upload" checkbox. The GUI includes "Upload selected '
            'params to FC" and "Skip" buttons at the bottom. When "Upload Selected to FC" is clicked, it uploads '
            'the selected parameters to the flight controller. When "Skip" is pressed, it skips to the next '
            "intermediate parameter file. The process gets repeated for each intermediate parameter file.",
            "No serial ports found",
            "Vehicle type not set explicitly, auto-detected %s.",
            "Vehicle type explicitly set to %s.",
            "Unknown",
            "ArduPilot Methodic Configurator Version: {_version}\n\n"
            "A clear configuration sequence for ArduPilot vehicles.\n\n"
            "Copyright © 2024-2025 Amilcar do Carmo Lucas and ArduPilot.org\n\n"
            "Licensed under the GNU General Public License v3.0",
        ]

        for expected in expected_missing:
            assert any(expected in norm_entry for norm_entry in normalized_entries), (
                f"Expected string not found in normalized entries: {expected}"
            )

    @patch("extract_missing_translations.logging.error")
    @patch("extract_missing_translations.open")
    @patch("extract_missing_translations.gettext.translation")
    def test_extract_missing_translations_with_malformed_line(self, mock_translation, mock_open, mock_logging) -> None:
        # Mock the gettext translation
        mock_translator = MagicMock()
        mock_translator.gettext.side_effect = lambda x: x
        mock_translation.return_value = mock_translator

        # Mock file reading with a malformed line
        mock_file = MagicMock()
        mock_file.__enter__.return_value.readlines.return_value = [
            "msgid bad format line\n",  # Malformed line
            'msgstr ""\n',
        ]
        mock_open.return_value = mock_file

        result = extract_missing_translations.extract_missing_translations("ja")
        # Check that the result is empty (malformed lines are skipped)
        assert len(result) == 0
        # Check that the error was logged
        mock_logging.assert_called_once()

    def test_output_to_files_with_empty_list(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdirname:
            output_path = os.path.join(tmpdirname, "test_empty")
            extract_missing_translations.output_to_files([], output_path, 10)
            # No files should be created for empty list
            assert len(os.listdir(tmpdirname)) == 0

    def test_output_to_files_exact_max_translations(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdirname:
            output_path = os.path.join(tmpdirname, "test_exact")
            missing_translations = [(1, "Test 1"), (2, "Test 2"), (3, "Test 3")]
            extract_missing_translations.output_to_files(missing_translations, output_path, 3)
            # Should create exactly one file
            assert len(os.listdir(tmpdirname)) == 1

            with open(f"{output_path}.txt", encoding="utf-8") as f:
                content = f.read()
                assert "1:Test 1" in content
                assert "2:Test 2" in content
                assert "3:Test 3" in content

    def test_output_to_files_multiple_files(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdirname:
            output_path = os.path.join(tmpdirname, "test_multiple")
            missing_translations = [(1, "Test 1"), (2, "Test 2"), (3, "Test 3"), (4, "Test 4"), (5, "Test 5")]
            extract_missing_translations.output_to_files(missing_translations, output_path, 2)
            # Should create 3 files (ceiling division of 5/2)
            assert len(os.listdir(tmpdirname)) == 3

            with open(f"{output_path}_1.txt", encoding="utf-8") as f:
                content = f.read()
                assert "1:Test 1" in content
                assert "2:Test 2" in content

            with open(f"{output_path}_2.txt", encoding="utf-8") as f:
                content = f.read()
                assert "3:Test 3" in content
                assert "4:Test 4" in content

            with open(f"{output_path}_3.txt", encoding="utf-8") as f:
                content = f.read()
                assert "5:Test 5" in content

    @patch("extract_missing_translations.glob.glob")
    @patch("extract_missing_translations.os.remove")
    def test_output_to_files_removes_existing_files(self, mock_remove, mock_glob) -> None:
        mock_glob.side_effect = lambda pattern: ["file1.txt", "file2.txt", "file3.txt"] if pattern.endswith("*.txt") else []

        extract_missing_translations.output_to_files([], "test_output", 10)

        # Should attempt to remove all three existing files
        assert mock_remove.call_count == 3
        calls = [call("file1.txt"), call("file2.txt"), call("file3.txt")]
        mock_remove.assert_has_calls(calls, any_order=True)

    @patch("extract_missing_translations.parse_arguments")
    @patch("extract_missing_translations.extract_missing_translations")
    @patch("extract_missing_translations.output_to_files")
    def test_main_with_large_number_of_translations(self, mock_output, mock_extract, mock_args) -> None:
        # Test main with a large number of translations
        mock_args_instance = MagicMock()
        mock_args_instance.lang_code = "fr"
        mock_args_instance.output_file = "test_large"
        mock_args_instance.max_translations = 5
        mock_args.return_value = mock_args_instance

        # Create a large list of translations
        large_translations = [(i, f"Test string {i}") for i in range(20)]
        mock_extract.return_value = large_translations

        extract_missing_translations.main()

        # Verify the function was called with the correct parameters
        mock_extract.assert_called_once_with("fr")
        mock_output.assert_called_once_with(large_translations, "test_large_fr", 5)

    @patch("extract_missing_translations.open")
    @patch("extract_missing_translations.gettext.translation")
    def test_extract_missing_translations(self, mock_translation, mock_open) -> None:
        # Mock the gettext translation
        mock_translator = MagicMock()

        # Make gettext return the same string to simulate missing translations for Test string 1
        # But return a different string for Test string 2 to simulate it's already translated
        def mock_gettext(text: str) -> str:
            if text == "Test string 1":
                return text
            return "Translated Test string 2"

        mock_translator.gettext.side_effect = mock_gettext
        mock_translation.return_value = mock_translator

        # Mock file reading
        mock_file = MagicMock()
        mock_file.__enter__.return_value.readlines.return_value = [
            'msgid "Test string 1"\n',
            'msgstr ""\n',
            'msgid "Test string 2"\n',
            'msgstr "Translated"\n',
            "\n",
        ]
        mock_open.return_value = mock_file

        result = extract_missing_translations.extract_missing_translations("ja")
        assert len(result) == 1
        assert result[0][1] == "Test string 1"

    @patch("extract_missing_translations.parse_arguments")
    @patch("extract_missing_translations.extract_missing_translations")
    @patch("extract_missing_translations.output_to_files")
    @patch("extract_missing_translations.logging.basicConfig")
    def test_main(self, mock_logging, mock_output, mock_extract, mock_args) -> None:
        # Test the main function
        mock_args_instance = MagicMock()
        mock_args_instance.lang_code = "ja"
        mock_args_instance.output_file = "test_output"
        mock_args_instance.max_translations = 50
        mock_args.return_value = mock_args_instance

        mock_extract.return_value = [(1, "Test string 1")]

        extract_missing_translations.main()

        mock_extract.assert_called_once_with("ja")
        mock_output.assert_called_once_with([(1, "Test string 1")], "test_output_ja", 50)
        mock_logging.assert_called_once()


# These are pytest-style standalone functions, outside the class
@pytest.fixture
def temp_dir() -> str:
    with tempfile.TemporaryDirectory() as tmpdirname:
        yield tmpdirname


def test_output_to_files_real(temp_dir) -> None:  # pylint: disable=redefined-outer-name
    # Test with real file output
    missing_translations = [(1, "Test string 1"), (5, "Test string 2"), (10, "Test string 3")]
    output_path = os.path.join(temp_dir, "test_output")

    extract_missing_translations.output_to_files(missing_translations, output_path, 2)

    # Check that files were created
    assert os.path.exists(f"{output_path}_1.txt")
    assert os.path.exists(f"{output_path}_2.txt")

    # Check file contents
    with open(f"{output_path}_1.txt", encoding="utf-8") as f:
        content = f.read()
        assert "1:Test string 1" in content
        assert "5:Test string 2" in content

    with open(f"{output_path}_2.txt", encoding="utf-8") as f:
        content = f.read()
        assert "10:Test string 3" in content


def test_extract_missing_translations_with_real_po_file() -> None:
    # Test with the actual test PO file to verify it works correctly
    test_po_file = os.path.join(os.path.dirname(__file__), "test_extract_missing_translations_input.po")

    # Skip this test if the file doesn't exist
    if not os.path.exists(test_po_file):
        pytest.skip(f"Test PO file not found: {test_po_file}")

    # Use a single with statement with multiple contexts instead of nested with statements
    with (
        patch("extract_missing_translations.os.path.join", return_value=test_po_file),
        patch("extract_missing_translations.gettext.translation") as mock_translation,
    ):
        # Mock translator to simulate all strings are untranslated
        mock_translator = MagicMock()
        mock_translator.gettext.side_effect = lambda x: x
        mock_translation.return_value = mock_translator

        result = extract_missing_translations.extract_missing_translations("ja")

        # Verify we have exactly 6 missing translations
        assert len(result) == 6

        # Create a dictionary of line numbers to strings for easier verification
        missing_by_line = dict(result)

        # Verify all expected strings are present with correct line numbers
        expected_strings = [
            (20, "ArduPilot methodic configurator is a simple GUI"),
            (24, "No serial ports found"),
            (27, "Vehicle type not set explicitly, auto-detected %s."),
            (32, "Vehicle type explicitly set to %s."),
            (45, "Unknown"),
            (56, "ArduPilot Methodic Configurator Version:"),
        ]

        for expected_line, expected_text in expected_strings:
            assert expected_line in missing_by_line
            assert expected_text in missing_by_line[expected_line]



"""
Check for software updates and install them if available.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar Lucas

SPDX-License-Identifier: GPL-3.0-or-later
"""

import os
import subprocess
import tempfile
import time
from datetime import datetime, timezone
from logging import debug as logging_debug
from logging import error as logging_error
from logging import info as logging_info
from logging import shutdown as logging_shutdown
from typing import Any, Callable, Optional
from urllib.parse import urljoin
from webbrowser import open as webbrowser_open

from requests import HTTPError as requests_HTTPError
from requests import RequestException as requests_RequestException
from requests import Timeout as requests_Timeout
from requests import get as requests_get
from requests.exceptions import RequestException

from ardupilot_methodic_configurator import _

# Constants
GITHUB_API_URL_RELEASES = "https://api.github.com/repos/ArduPilot/MethodicConfigurator/releases/"


def download_file_from_url(
    url: str, local_filename: str, timeout: int = 30, progress_callback: Optional[Callable[[float, str], None]] = None
) -> bool:
    if not url or not local_filename:
        logging_error(_("URL or local filename not provided."))
        return False

    logging_info(_("Downloading %s from %s"), local_filename, url)

    try:
        proxies_dict = {
            "http": os.environ.get("HTTP_PROXY") or os.environ.get("http_proxy"),
            "https": os.environ.get("HTTPS_PROXY") or os.environ.get("https_proxy"),
            "no_proxy": os.environ.get("NO_PROXY") or os.environ.get("no_proxy"),
        }

        # Remove None values
        proxies = {k: v for k, v in proxies_dict.items() if v is not None}

        # Make request with proxy support
        response = requests_get(
            url,
            stream=True,
            timeout=timeout,
            proxies=proxies,
            verify=True,  # SSL verification
        )
        response.raise_for_status()

        total_size = int(response.headers.get("content-length", 0))
        block_size = 8192
        downloaded = 0

        os.makedirs(os.path.dirname(os.path.abspath(local_filename)), exist_ok=True)

        with open(local_filename, "wb") as file:
            for chunk in response.iter_content(chunk_size=block_size):
                if chunk:
                    file.write(chunk)
                    downloaded += len(chunk)
                    if progress_callback and total_size:
                        progress = (downloaded / total_size) * 100
                        msg = _("Downloading ... {:.1f}%")
                        progress_callback(progress, msg.format(progress))

        if progress_callback:
            progress_callback(100.0, _("Download complete"))
        return bool(downloaded > 0)

    except requests_Timeout:
        logging_error(_("Download timed out"))
    except requests_RequestException as e:
        logging_error(_("Network error during download: {}").format(e))
    except OSError as e:
        logging_error(_("File system error: {}").format(e))
    except ValueError as e:
        logging_error(_("Invalid data received from %s: %s"), url, e)

    return False


def get_release_info(name: str, should_be_pre_release: bool, timeout: int = 30) -> dict[str, Any]:
    """
    Get release information from GitHub API.

    Args:
        name: Release name/path (e.g. '/latest')
        should_be_pre_release: Whether the release should be a pre-release
        timeout: Request timeout in seconds

    Returns:
        Release information dictionary

    Raises:
        RequestException: If the request fails

    """
    if not name:
        msg = "Release name cannot be empty"
        raise ValueError(msg)

    try:
        url = urljoin(GITHUB_API_URL_RELEASES, name.lstrip("/"))
        response = requests_get(url, timeout=timeout)
        response.raise_for_status()

        release_info = response.json()

        if should_be_pre_release and not release_info["prerelease"]:
            logging_error(_("The latest continuous delivery build must be a pre-release"))
        if not should_be_pre_release and release_info["prerelease"]:
            logging_error(_("The latest stable release must not be a pre-release"))

        return release_info  # type: ignore[no-any-return]

    except requests_HTTPError as e:
        if e.response.status_code == 403:
            logging_error(_("Failed to fetch release info: {}").format(e))
            # Get the rate limit reset time
            reset_time = int(e.response.headers.get("X-RateLimit-Reset", 0))
            # Create a timezone-aware UTC datetime
            reset_datetime = datetime.fromtimestamp(reset_time, timezone.utc).strftime("%Y-%m-%d %H:%M:%S %Z")
            logging_error(_("Rate limit exceeded. Please try again after: %s (UTC)"), reset_datetime)
        raise
    except RequestException as e:
        logging_error(_("Failed to fetch release info: {}").format(e))
        raise
    except (KeyError, ValueError) as e:
        logging_error(_("Invalid release data: {}").format(e))
        raise


def download_and_install_on_windows(
    download_url: str,
    file_name: str,
    progress_callback: Optional[Callable[[float, str], None]] = None,
) -> bool:
    """
    Download and install a new version of the application on Windows.

    This function handles the complete update process:
    1. Downloads the installer from the provided URL
    2. Creates a batch script to run the installer after the current process exits
    3. Exits the current application to allow the installer to run without conflicts
       (prevents "program is already running" errors during installation)

    Args:
        download_url: The URL from which to download the installer
        file_name: The name to save the downloaded file as
        progress_callback: Optional callback function to report progress
                          Takes two arguments: progress (0.0-1.0) and status message

    Returns:
        bool: True if the process started successfully (note: if successful,
              the application will exit and never return from this function)

    """
    logging_info(_("Downloading and installing new version for Windows..."))
    try:
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = os.path.join(temp_dir, file_name)

            # Download with progress updates
            if not download_file_from_url(
                download_url,
                temp_path,
                timeout=60,  # Increased timeout for large files
                progress_callback=progress_callback,
            ):
                logging_error(_("Failed to download installer from %s"), download_url)
                return False

            if progress_callback:
                progress_callback(0.0, _("Starting installation..."))

            # Create a batch file to run the installer after this process exits
            batch_file_path = os.path.join(temp_dir, "run_installer.bat")
            with open(batch_file_path, "w", encoding="utf-8") as batch_file:
                # Wait a moment for the main process to exit
                batch_file.write("@echo off\n")
                batch_file.write("ping 127.0.0.1 -n 2 > nul\n")  # Wait ~1 second
                batch_file.write(f'if exist "{temp_path}" (\n')  # Check if installer still exists
                batch_file.write(f'  start "" "{temp_path}"\n')  # Run the installer
                batch_file.write(") else (\n")
                batch_file.write("  echo Installer not found\n")
                batch_file.write(")\n")
                batch_file.write("del %0\n")  # Delete the batch file itself

            # Make the batch file executable and start it
            # Use 'with' pattern for the Popen operation to fix R1732
            with subprocess.Popen(  # noqa: S602
                [batch_file_path],
                shell=True,
                creationflags=subprocess.CREATE_NO_WINDOW | subprocess.DETACHED_PROCESS,  # type: ignore[attr-defined]
            ) as _result:
                pass  # We don't need to do anything with the process object

            # Signal to the main application that it should exit
            if progress_callback:
                progress_callback(1.0, _("Installation ready. Application will restart now."))

            # Ensure logs are flushed before exit
            logging_shutdown()

            # Give a moment for the callback to complete
            time.sleep(0.5)

            # Exit this process to allow the installer to run
            # This means the function will never actually return at this point,
            # but we need to maintain consistent return type for the function signature
            os._exit(0)  # Force exit without cleanup

    except subprocess.SubprocessError as e:
        logging_error(_("Installation failed: {}").format(e))
    except OSError as e:
        logging_error(_("File operation failed: {}").format(e))
    return False


def download_and_install_pip_release(progress_callback: Optional[Callable[[float, str], None]] = None) -> int:
    logging_info(_("Updating via pip for Linux and macOS..."))

    if progress_callback:
        progress_callback(0.0, _("Starting installation..."))

    ret = os.system("pip install --upgrade ardupilot_methodic_configurator")  # noqa: S605, S607

    if ret == 0 and progress_callback:
        progress_callback(100.0, _("Download complete"))

    return ret


def verify_and_open_url(url: str) -> bool:
    """
    Verify if a URL is accessible and open it in the default web browser if successful.

    Args:
        url: The URL to verify and open

    Returns:
        bool: True if the URL was found and opened, False otherwise

    """
    if not url:
        logging_error(_("URL not provided."))
        return False

    logging_debug(_("Verifying URL: %s"), url)
    url_found: bool = False

    try:
        # Get proxy settings from environment variables
        proxies_dict = {
            "http": os.environ.get("HTTP_PROXY") or os.environ.get("http_proxy"),
            "https": os.environ.get("HTTPS_PROXY") or os.environ.get("https_proxy"),
            "no_proxy": os.environ.get("NO_PROXY") or os.environ.get("no_proxy"),
        }
        # Remove None values
        proxies = {k: v for k, v in proxies_dict.items() if v is not None}

        # Use requests.get with allow_redirects to handle HTTP redirects properly
        response = requests_get(url, timeout=5, stream=True, allow_redirects=True, proxies=proxies, verify=True)
        url_found = response.status_code == 200
    except (requests_RequestException, requests_Timeout) as e:
        logging_error(_("Failed to access URL: %s. Error: %s"), url, str(e))
        url_found = False

    if url_found:
        try:
            logging_info(_("Opening URL in browser: %s"), url)
            webbrowser_open(url=url, new=0, autoraise=True)
        except Exception as e:  # pylint: disable=broad-exception-caught
            logging_error(_("Failed to open URL in browser: %s"), str(e))
            url_found = False

    return url_found



"""
Python package initialization file. Loads translations and declares version information.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from ardupilot_methodic_configurator.internationalization import load_translation

_ = load_translation()

__version__ = "1.4.5"
